from __future__ import annotations
from base64 import b64decode
from musig.abstractclasses import AbstractSingleSigKey
from musig.nonce import Nonce
from musig.partialsignature import PartialSignature
from musig.publickey import PublicKey
from musig.signature import Signature
from nacl.signing import SigningKey, VerifyKey
import nacl.bindings


class SingleSigKey(AbstractSingleSigKey):
    """A simple-to-use class that generates 1-of-1 MuSig signatures."""

    def __init__(self, data: dict = None) -> None:
        """Initialize using a nacl.signing.SigningKey or deserialize.
            Call with `{'skey':SigningKey}` to create a new SingleSigKey.
            Call with a dict containing json.loads output from a json.dumps
            serialization to restore an instance.
        """
        if data is None:
            raise ValueError('cannot instantiate an empty SingleSigKey')

        if not isinstance(data, dict):
            raise TypeError('data for initialization must be of type dict')

        if 'skey' in data:
            if isinstance(data['skey'], SigningKey):
                self.skey = data['skey']
            else:
                seed = data['skey'] if isinstance(data['skey'], bytes) else b64decode(data['skey'])
                self.skey = SigningKey(seed)

        if 'vkey_base' in data:
            if isinstance(data['vkey_base'], VerifyKey):
                self.vkey_base = data['vkey_base']
            else:
                vkey_base = data['vkey_base'] if type(data['vkey_base']) is bytes else b64decode(data['vkey_base'])
                self.vkey_base = VerifyKey(vkey_base)

        if 'vkey' in data:
            if isinstance(data['vkey'], PublicKey):
                self.vkey = data['vkey']
            else:
                vkey = data['vkey'] if type(data['vkey']) is bytes else b64decode(data['vkey'])
                self.vkey = PublicKey.from_bytes(vkey)

        if self.vkey_base is None and self.skey is not None:
            self.vkey_base = self.skey.verify_key

        if self.vkey is None and self.vkey_base is not None:
            self.vkey = PublicKey.create([self.vkey_base])

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on an instance; i.e. serialize to bytes."""
        if self.skey is not None:
            return bytes(self.skey)
        return b''

    @classmethod
    def from_bytes(cls, data: bytes) -> SingleSigKey:
        """Deserializes output from __bytes__."""
        if type(data) is not bytes:
            raise TypeError('bytes input must have length of ' +
                    str(nacl.bindings.crypto_scalarmult_ed25519_SCALARBYTES))
        if len(data) != nacl.bindings.crypto_scalarmult_ed25519_SCALARBYTES:
            raise ValueError('bytes input must have length of ' +
                str(nacl.bindings.crypto_scalarmult_ed25519_SCALARBYTES))

        return cls({'skey': SigningKey(data)})

    def sign_message(self, M: bytes) -> Signature:
        """Sign a message (M) and return a Signature."""
        nonce = Nonce()
        sig = PartialSignature.create(self.skey, nonce.r, self.vkey.L,
            self.vkey, nonce.R, M)
        return Signature.create(nonce.R, M, [sig])

    @property
    def skey(self) -> SigningKey|None:
        """The SigningKey used for creating signatures."""
        return self._skey if hasattr(self, '_skey') else None

    @skey.setter
    def skey(self, data: SigningKey):
        """The SigningKey used for creating signatures."""
        if not isinstance(data, SigningKey):
            raise TypeError('skey must be SigningKey')

        self['skey'] = data

    @property
    def vkey(self) -> PublicKey|None:
        """The aggregate public key for verifying signatures."""
        return self._vkey if hasattr(self, '_vkey') else None

    @vkey.setter
    def vkey(self, data: PublicKey):
        """The aggregate public key for verifying signatures."""
        if not isinstance(data, PublicKey):
            raise TypeError('vkey must be PublicKey')

        self['vkey'] = data

    @property
    def vkey_base(self) -> VerifyKey|None:
        """The VerifyKey base used to calculate the aggregate public key."""
        return self._vkey_base if hasattr(self, '_vkey_base') else None

    @vkey_base.setter
    def vkey_base(self, data: VerifyKey):
        """The VerifyKey base used to calculate the aggregate public key."""
        if not isinstance(data, VerifyKey):
            raise TypeError('vkey_base must be VerifyKey')

        self['vkey_base'] = data



from context import musig
from nacl.signing import SigningKey
import unittest


class TestMuSigHelpers(unittest.TestCase):
    """Test suite for the helper functions used by the various classes."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]

    def setUp(self):
        pass

    def test_clamp_scalar_raises_ValueError_when_given_invalid_arg(self):
        with self.assertRaises(ValueError):
            musig.clamp_scalar(None)

    def test_clamp_scalar_returns_bytes_different_from_unclamped_arg(self):
        original = bytes.fromhex('ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff')
        clamped = musig.clamp_scalar(original)
        assert type(clamped) is bytes
        assert original != clamped

    def test_aggregate_points_raises_TypeError_for_invalid_args(self):
        with self.assertRaises(TypeError):
            musig.aggregate_points(['123', None])

    def test_aggregate_points_raises_ValueError_for_invalid_points(self):
        with self.assertRaises(ValueError):
            musig.aggregate_points([bytes.fromhex(''.join(['ff' for i in range(32)]))])

        with self.assertRaises(ValueError):
            musig.aggregate_points([bytes.fromhex(''.join(['00' for i in range(32)]))])

    def test_aggregate_points_returns_bytes_when_given_VerifyKey_args(self):
        points = self.verify_keys
        assert type(musig.aggregate_points(points)) is bytes

    def test_H_big_returns_bytes(self):
        assert type(musig.H_big(b'asd', b'sdsd')) is bytes

    def test_H_small_returns_bytes(self):
        assert type(musig.H_small(b'asd', b'dfdfd')) is bytes

    def test_derive_key_from_seed_returns_bytes(self):
        assert type(musig.derive_key_from_seed(b'sds')) is bytes

    def test_derive_challenge_returns_bytes(self):
        L = b'123'
        X_i = b'321'
        X = b'213'
        R = b'dsd'
        M = b'dsds'
        assert type(musig.derive_challenge(L, X_i, X, R, M)) is bytes

    def test_H_agg_returns_clamped_bytes(self):
        output = musig.H_agg(b'asds', b'adsd')
        assert type(output) is bytes
        assert musig.clamp_scalar(output) == output

    def test_H_sig_returns_clamped_bytes(self):
        output = musig.H_sig(b'adssd', b'asdsd', b'asdsdsd')
        assert type(output) is bytes
        assert musig.clamp_scalar(output) == output

    def test_xor_returns_bytes_and_xors_arguments(self):
        assert type(musig.xor(b'123', b'123')) is bytes
        assert musig.xor(b'1', b'1')[0] == 0

    def test_bytes_are_same_returns_bool_and_is_accurate(self):
        assert type(musig.bytes_are_same(b'1', b'1')) is bool
        assert musig.bytes_are_same(b'1', b'1')
        assert not musig.bytes_are_same(b'1', b'2')


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from base64 import b64encode, b64decode
from musig.abstractclasses import AbstractNonce
from musig.helpers import aggregate_points, clamp_scalar
from secrets import token_bytes
import nacl.bindings


class Nonce(AbstractNonce):
    """A class that handles generating, serializing, and deserializing nonces."""

    def __init__(self, data: dict = None) -> None:
        """Initialize the instance, either with supplied data or with new values.
            Call with `data=None` to create a new Nonce.
            Call with `data={r:b64val}` or `{r:b64val, R:b64val}` to restore a full Nonce.
            Call with `data={R:b64val}` to restore a public Nonce.
        """
        if not isinstance(data, dict) and data is not None:
            raise TypeError('Nonce can be instantiated only with None or dict.')

        if data is None:
            # create new nonce
            self.r = clamp_scalar(token_bytes(nacl.bindings.crypto_scalarmult_ed25519_SCALARBYTES))
            self.R = nacl.bindings.crypto_scalarmult_ed25519_base_noclamp(self.r)
        else:
            # restore a Nonce from dict (e.g. json.loads result)
            if 'R' in data:
                R = data['R']
                self.R = R if type(R) is bytes else b64decode(R)
            if 'r' in data:
                r = data['r']
                self.r = r if type(r) is bytes else b64decode(r)
                self.R = nacl.bindings.crypto_scalarmult_ed25519_base_noclamp(self.r)

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on an instance."""
        if self.r is not None:
            return b'r' + self.r
        if self.R is not None:
            return self.R
        return b''

    @classmethod
    def from_bytes(cls, data: bytes) -> Nonce:
        """Deserializes output from __bytes__."""
        if type(data) is not bytes:
            raise TypeError('cannot call from_bytes with non-bytes param')
        if len(data) not in (32, 33):
            raise ValueError('byte length must be 32 or 33')

        if len(data) == 33:
            # restore full Nonce with private scalar and public point.
            r = data[1:]
            R = nacl.bindings.crypto_scalarmult_ed25519_base_noclamp(r)
            return cls({
                'r': r,
                'R': R
            })
        else:
            # restore partial Nonce with just the public point.
            return cls({'R': data})

    def __add__(self, other: Nonce) -> Nonce:
        """Result of the + operation between two Nonces."""
        R_sum = aggregate_points([self.R, other.R])
        return Nonce.from_bytes(R_sum)

    def copy(self) -> Nonce:
        """Make a copy without serializing and deserializing."""
        return Nonce({**self})

    def public(self) -> Nonce:
        """Return a Nonce with only the public nonce point."""
        return self.__class__({'R': b64encode(self.R).decode()})

    # properties
    @property
    def r(self) -> bytes|None:
        """The private scalar value."""
        return self._r if hasattr(self, '_r') else None

    @r.setter
    def r(self, value: bytes):
        """The private scalar value."""
        if not isinstance(value, bytes):
            raise TypeError('r value must be bytes')
        if len(value) != 32:
            raise ValueError('r value must have length 32')

        self['r'] = value

    @property
    def R(self) -> bytes|None:
        """The public point value."""
        return self._R if hasattr(self, '_R') else None

    @R.setter
    def R(self, value: bytes):
        """The public point value."""
        if not isinstance(value, bytes):
            raise TypeError('R value must be bytes')
        if len(value) != 32:
            raise ValueError('R value must have length 32')

        self['R'] = value



from __future__ import annotations
from base64 import b64decode
from enum import Enum
from math import ceil, log2
from musig import Nonce, NonceCommitment, PartialSignature, Signature
from musig.abstractclasses import AbstractProtocolMessage
from musig.helpers import bytes_are_same
from nacl.signing import SigningKey, VerifyKey, SignedMessage as NaclSignedMessage
from uuid import UUID


class ProtocolState(Enum):
    """An enum containing all of the byte representations of possible protocol
        states as encoded in ProtocolMessage or referenced in SigningSession.
    """
    EMPTY = 0x00
    INITIALIZED = 0x10

    AWAITING_PARTICIPANT_KEY = 0x11
    SENDING_PARTICIPANT_KEY = 0x12
    ACK_PARTICIPANT_KEY = 0x13
    REJECT_PARTICIPANT_KEY = 0x1f

    AWAITING_COMMITMENT = 0x20
    SENDING_COMMITMENT = 0x21
    ACK_COMMITMENT = 0x22
    TIME_EXCEEDED_AWAITING_COMMITMENT = 0x2e
    REJECT_COMMITMENT = 0x2f

    AWAITING_MESSAGE = 0x30
    SENDING_MESSAGE = 0x31
    ACK_MESSAGE = 0x32
    REJECT_MESSAGE = 0x3f

    AWAITING_NONCE = 0x40
    SENDING_NONCE = 0x41
    ACK_NONCE = 0x42
    TIME_EXCEEDED_AWAITING_NONCE = 0x4e
    REJECT_NONCE = 0x4f

    AWAITING_PARTIAL_SIGNATURE = 0x50
    SENDING_PARTIAL_SIGNATURE = 0x51
    ACK_PARTIAL_SIGNATURE = 0x52
    TIME_EXCEEDED_AWAITING_PARTIAL_SIGNATURE = 0x5e
    REJECT_PARTIAL_SIGNATURE = 0x5f

    COMPLETED = 0xe0
    ABORTED = 0xff


class ProtocolError(Exception):
    """A custom error to be thrown if the SigningSession is configured to throw them."""
    def __init__(self, message: str, protocol_state: ProtocolState) -> None:
        self.message = message
        self.protocol_state = protocol_state
        super().__init__(message)

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on instance; i.e. serialize to bytes."""
        return self.protocol_state.value.to_bytes(1, 'little') + bytes(self.message, 'utf-8')

    def __str__(self) -> str:
        """Result of calling str() on an instance."""
        return self.protocol_state.name + ':' + self.message

    def __eq__(self, other) -> bool:
        """Timing-attack safe comparison."""
        if not isinstance(other, self.__class__):
            return False
        return bytes_are_same(bytes(self), bytes(other))

    def __hash__(self) -> int:
        """Result of calling hash() on instance; allows inclusion in sets and
            use as a key in a dict (of dubious worth, but technically possible.)
        """
        return hash(bytes(self))

    @classmethod
    def from_bytes(cls, bts: bytes) -> ProtocolError:
        """Deserializes output from __bytes__."""
        protocol_state = ProtocolState(bts[0])
        message = str(bts[1:], 'utf-8')
        return cls(message, protocol_state)

    @classmethod
    def from_str(cls, s: str) -> ProtocolError:
        """Deserializes output from __str__."""
        parts = s.split(':')
        protocol_state = ProtocolState[parts[0]]
        message = ':'.join(parts[1:])
        return cls(message, protocol_state)


class ProtocolMessage(AbstractProtocolMessage):
    """A class that handles packing and unpacking messages for mediating the
        protocol. This should be sufficient, but any other system can be used as
        long as the SigningSession methods are called in the right manner and
        the right data is communicated between parties in the right order.
        Multiple serialization options are available for convenience.
    """
    def __init__(self, data: dict = None) -> None:
        """Initialize with json.loads output of json.dumps serialization to
            restore an instance. Otherwise, it is better to use the create method.
        """
        if data is None:
            raise ValueError('cannot initialize an empty ProtocolMessage')
        if not isinstance(data, dict):
            raise TypeError('cannot initialize with non-dict data')
        if 'state' not in data or 'message' not in data:
            raise ValueError('must include a valid state and message to initialize')

        # parse session_id
        if 'session_id' in data:
            session_id = data['session_id'] if type(data['session_id']) is bytes else b64decode(data['session_id'])
            self.session_id = UUID(bytes=session_id)

        # parse state
        self.state = ProtocolState[data['state']]

        # parse message
        self.message = data['message'] if type(data['message']) is bytes else b64decode(data['message'])
        self.parse_message()

        # parse signature if present
        if 'signature' in data:
            sig = data['signature'] if type(data['signature']) is bytes else b64decode(data['signature'])
            sig = sig[:64]
            msg = self.state.value.to_bytes(1, 'little') + self.message
            self.signature = NaclSignedMessage._from_parts(sig, msg, sig+msg)

        # parse vkey if present
        if 'vkey' in data:
            self.vkey = VerifyKey(data['vkey'] if type(data['vkey']) is bytes else b64decode(data['vkey']))

    def __bytes__(self) -> bytes:
        """Serialize to bytes using custom (but simple) serialization scheme."""
        bta = bytearray()

        # first serializse the state
        bta.append(self.state.value)

        # next serialize the session_id
        if self.state is not ProtocolState.EMPTY:
            bta.extend(self.session_id.bytes)

        # serialize each message element to bytes
        m = self.message if type(self.message) is bytes else bytes(self.message)
        # only serialize if it has a length
        if len(m) > 0:
            # code for message
            bta.extend(b'm')
            # length of the message length
            m_len_len = ceil(log2(len(m) + 1) / 8)
            bta.append(m_len_len)
            # message length
            m_len = len(m).to_bytes(m_len_len, 'little')
            bta.extend(m_len)
            # message
            bta.extend(m)

        # encode the signature if it is set
        if self.signature is not None:
            # code for signature
            bta.extend(b's')
            # byte encoding of signature always has len 64
            bta.extend(self.signature.signature)

        # encode the vkey if it is set
        if self.vkey is not None:
            # code for vkey
            bta.extend(b'v')
            # byte encoding of vkey is always 32 bytes
            bta.extend(bytes(self.vkey))

        return bytes(bta)

    @classmethod
    def from_bytes(cls, data: bytes) -> ProtocolMessage:
        """Deserialize from bytes using custom (but simple) serialization scheme."""
        if type(data) not in (bytes, bytearray):
            raise TypeError('supplied data must be of type bytes or bytearray')

        # first parse the protocol state
        state = ProtocolState(data[0])
        data = data[1:]

        message = b''
        session_id = None
        signature = None
        vkey = None

        # now the session_id if applicable
        if state is not ProtocolState.EMPTY:
            session_id = data[:16]
            data = data[16:]

        # parse each section of the message
        while len(data) > 0:
            code = data[0:1]
            data = data[1:]

            # parse a message
            if code == b'm':
                # length of message length
                m_len_len = data[0]
                # message length
                m_len = int.from_bytes(data[1:1+m_len_len], 'little')
                # message
                m = data[1+m_len_len:1+m_len_len+m_len]
                message = m
                # remove from data
                data = data[1+m_len_len+m_len:]

            # parse a signature
            if code == b's':
                # signature is always 64 bytes
                signature = data[:64]
                data = data[64:]

            # parse a vkey
            if code == b'v':
                # vkey is always 32 bytes
                vkey = data[:32]
                data = data[32:]

        # instantiate
        param = {
            'state': state.name,
            'message': message
        }

        if session_id is not None:
            param['session_id'] = session_id

        if signature is not None:
            param['signature'] = signature

        if vkey is not None:
            param['vkey'] = vkey

        return cls(param)

    def parse_message(self) -> None:
        """Parses the message into parts based upon protocol state, e.g. lists
            of participant keys, nonces, nonce commitments, etc, storing the
            result in self.message_parts/self['message_parts'].
        """
        parts = []
        message = self.message
        if self.state in (
                            ProtocolState.SENDING_PARTICIPANT_KEY,
                            ProtocolState.ACK_PARTICIPANT_KEY,
                            ProtocolState.REJECT_PARTICIPANT_KEY,
                            ProtocolState.AWAITING_COMMITMENT,
                            ProtocolState.AWAITING_NONCE,
                            ProtocolState.AWAITING_PARTIAL_SIGNATURE,
                        ):
            # every 32 bytes will be a VerifyKey
            while len(message) > 0:
                parts.append(VerifyKey(message[:32]))
                message = message[32:]

        if self.state in (
                            ProtocolState.SENDING_COMMITMENT,
                            ProtocolState.ACK_COMMITMENT,
                            ProtocolState.REJECT_COMMITMENT,
                        ):
            # every 32 bytes will be a NonceCommitment
            while len(message) > 0:
                parts.append(NonceCommitment.from_bytes(message[:32]))
                message = message[32:]

        if self.state in (
                            ProtocolState.SENDING_NONCE,
                            ProtocolState.ACK_NONCE,
                            ProtocolState.REJECT_NONCE,
                        ):
            # every 32 bytes will be a Nonce
            while len(message) > 0:
                parts.append(Nonce.from_bytes(message[:32]))
                message = message[32:]

        if self.state in (
                            ProtocolState.SENDING_PARTIAL_SIGNATURE,
                            ProtocolState.ACK_PARTIAL_SIGNATURE,
                            ProtocolState.REJECT_PARTIAL_SIGNATURE,
                        ):
            # every 32 bytes will be a PartialSignature
            while len(message) > 0:
                parts.append(PartialSignature.from_bytes(message[:32]))
                message = message[32:]

        if self.state is ProtocolState.COMPLETED:
            parts = [Signature.from_bytes(message)]

        if self.state is ProtocolState.ABORTED:
            parts = [ProtocolError.from_bytes(message)]

        self.message_parts = parts

    def add_signature(self, skey: SigningKey) -> ProtocolMessage:
        """Add a signature to the message using the provided SigningKey. This
            signs the protocol state + message, attaching the signature and the
            relevant VerifyKey to the instance.
        """
        self.signature = skey.sign(self.state.value.to_bytes(1, 'little') + self.message)
        self.vkey = skey.verify_key
        return self

    def check_signature(self) -> bool:
        """Return true if and only if the instance contains a VerifyKey and a
            signature valid for the protocol state + message, else return False.
        """
        try:
            if self.signature is None or self.vkey is None:
                return False
            self.vkey.verify(self.signature)
            return True
        except:
            return False

    @classmethod
    def create(cls, id: UUID|None, state: ProtocolState, data: list) -> ProtocolMessage:
        """Create a new instance with the given id, state, and data.
            If the state is EMPTY, the id is ignored.

            If the state is SENDING_PARTICIPANT_KEY, ACK_PARTICIPANT_KEY,
            REJECT_PARTICIPANT_KEY, AWAITING_COMMITMENT, AWAITING_NONCE, or
            AWAITING_PARTIAL_SIGNATURE, the data must be a list of
            VerifyKey|bytes.

            If the state is SENDING_COMMITMENT, ACK_COMMITMENT, or
            REJECT_COMMITMENT, the data must be a list of NonceCommitment|bytes.

            If the state is SENDING_MESSAGE, ACK_MESSAGE, or REJECT_MESSAGE, the
            data must be a list of str|bytes.

            If the state is SENDING_NONCE, ACK_NONCE, or REJECT_NONCE, the data
            must be a list of Nonce|bytes.

            If the state is SENDING_PARTIAL_SIGNATURE, ACK_PARTIAL_SIGNATURE, or
            REJECT_PARTIAL_SIGNATURE, the data must be a list of
            PartialSignature|bytes.

            If the state is COMPLETED, the data must be a Signature|bytes.

            If the state is ABORTED, the data must be a ProtocolError|bytes.
        """
        if not isinstance(id, UUID) and state is not ProtocolState.EMPTY:
            raise TypeError('id must be a valid UUID for non-EMPTY state')

        if not isinstance(state, ProtocolState):
            raise TypeError('state must be a valid ProtocolState')

        if type(data) not in (list, tuple):
            raise TypeError('data must be a list or tuple of values')

        if state is ProtocolState.EMPTY:
            return cls({
                'state': state.name,
                'message': ''
            })

        if state in (
                        ProtocolState.INITIALIZED,
                        ProtocolState.AWAITING_MESSAGE,
                        ProtocolState.AWAITING_PARTICIPANT_KEY,
                    ):
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': ''
            })

        if state in (
                        ProtocolState.SENDING_PARTICIPANT_KEY,
                        ProtocolState.ACK_PARTICIPANT_KEY,
                        ProtocolState.REJECT_PARTICIPANT_KEY,
                        ProtocolState.AWAITING_COMMITMENT,
                        ProtocolState.AWAITING_NONCE,
                        ProtocolState.AWAITING_PARTIAL_SIGNATURE,
                    ):
            for vk in data:
                if not isinstance(vk, VerifyKey) and (type(vk) is not bytes or len(vk) != 32):
                    raise TypeError(f'each datum for ProtocolMessage with state={state.name} must be VerifyKey or 32 bytes')
            vkeys = [vk if type(vk) is bytes else bytes(vk) for vk in data]
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': b''.join(vkeys)
            })

        if state in (
                        ProtocolState.SENDING_COMMITMENT,
                        ProtocolState.ACK_COMMITMENT,
                        ProtocolState.REJECT_COMMITMENT,
                    ):
            for nc in data:
                if not isinstance(nc, NonceCommitment) and (type(nc) is not bytes or len(nc) != 32):
                    raise TypeError(f'each datum for ProtocolMessage with state={state.name} must be NonceCommitment or 32 bytes')
            commitments = [nc if type(nc) is bytes else bytes(nc) for nc in data]
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': b''.join(commitments)
            })

        if state in (
                        ProtocolState.SENDING_MESSAGE,
                        ProtocolState.ACK_MESSAGE,
                        ProtocolState.REJECT_MESSAGE,
                    ):
            if not type(data[0]) in (str, bytes):
                raise TypeError(f'datum for ProtocolMessage with state={state.name} must be str or bytes')
            msg = data[0] if type(data[0]) is bytes else bytes(data[0])
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': msg
            })

        if state in (
                        ProtocolState.SENDING_NONCE,
                        ProtocolState.ACK_NONCE,
                        ProtocolState.REJECT_NONCE,
                    ):
            for n in data:
                if not isinstance(n, Nonce) and (type(n) is not bytes or len(n) != 32):
                    raise TypeError(f'each datum for ProtocolMessage with state={state.name} must be Nonce or 32 bytes')
            nonces = [n if type(n) is bytes else bytes(n) for n in data]
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': b''.join(nonces)
            })

        if state in (
                        ProtocolState.SENDING_PARTIAL_SIGNATURE,
                        ProtocolState.ACK_PARTIAL_SIGNATURE,
                        ProtocolState.REJECT_PARTIAL_SIGNATURE,
                    ):
            for s in data:
                if not isinstance(s, PartialSignature) and (type(s) is not bytes or len(s) != 32):
                    raise TypeError(f'each datum for ProtocolMessage with state={state.name} must be PartialSignature or 32 bytes')
            sigs = [s if type(s) is bytes else bytes(s.public()) for s in data]
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': b''.join(sigs)
            })

        if state is ProtocolState.COMPLETED:
            if not isinstance(data[0], Signature) and (type(data[0]) is not bytes or len(data[0]) < 65):
                raise TypeError(f'datum for ProtocolMessage with state={state.name} must be Siganture or > 64 bytes')
            sig = data[0] if type(data[0]) is bytes else bytes(data[0])
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': bytes(sig)
            })

        if state is ProtocolState.ABORTED:
            if not isinstance(data[0], ProtocolError) and (type(data[0]) is not bytes or len(data[0]) < 1):
                raise TypeError(f'each datum for ProtocolMessage with state={state.name} must be ProtocolError or >0 bytes')
            err = data[0] if type(data[0]) is bytes else bytes(data[0])
            return cls({
                'session_id': id.bytes,
                'state': state.name,
                'message': err
            })

    # properties
    @property
    def session_id(self) -> UUID|None:
        """The UUID of the signing session for which this message was constructed."""
        return self._session_id if hasattr(self, '_session_id') else None

    @session_id.setter
    def session_id(self, data: UUID):
        """The UUID of the signing session for which this message was constructed."""
        if type(data) not in (UUID, bytes):
            raise TypeError('session_id must be UUID or bytes')

        self['session_id'] = data if type(data) is bytes else data.bytes
        self._session_id = UUID(bytes=data) if type(data) is bytes else data

    @property
    def state(self) -> ProtocolState:
        """The protocol state of the message."""
        return self._state if hasattr(self, '_state') else None

    @state.setter
    def state(self, data: ProtocolState):
        """The protocol state of the message."""
        if type(data) is not ProtocolState:
            raise TypeError('state must be ProtocolState')

        self['state'] = data.name
        self._state = data

    @property
    def message(self) -> bytes:
        """The message itself."""
        return self._message if hasattr(self, '_message') else None

    @message.setter
    def message(self, data: bytes):
        """The message itself."""
        if type(data) not in (bytes, str):
            raise TypeError('message must be bytes or str')

        self['message'] = data if type(data) is bytes else bytes(data, 'utf-8')

    @property
    def message_parts(self) -> tuple:
        """The things that serialize into and deserialize from self.message."""
        return self._message_parts if hasattr(self, '_message_parts') else tuple()

    @message_parts.setter
    def message_parts(self, data: tuple):
        """The things that serialize into and deserialize from self.message."""
        if type(data) not in (tuple, list):
            raise TypeError('message_parts must be list or tuple of values')

        self['message_parts'] = tuple(data)

    @property
    def signature(self) -> NaclSignedMessage|None:
        """The Signature result of add_signature."""
        return self._signature if hasattr(self, '_signature') else None

    @signature.setter
    def signature(self, data: NaclSignedMessage):
        """The Signature result of add_signature."""
        if type(data) is not NaclSignedMessage:
            raise TypeError('signature must be nacl.signing.SignedMessage')

        self['signature'] = data

    @property
    def vkey(self) -> VerifyKey|None:
        """The VerifyKey used to verify the signature."""
        return self._vkey if hasattr(self, '_vkey') else None

    @vkey.setter
    def vkey(self, data: VerifyKey):
        """The VerifyKey used to verify the signature."""
        if type(data) is not VerifyKey:
            raise TypeError('vkey must be VerifyKey')

        self['vkey'] = data



"""Constants used for protecting against various types of Wagner/generalized
    birthday attacks. If a protocol step is not completed before the specified
    time has been passed, the protocol will abort and raise an error.
"""


MAX_WAIT_TIME_FOR_COMMITMENTS = 3600 # seconds; i.e. 1 hour
MAX_WAIT_TIME_FOR_PUBLIC_NONCES = 600 # seconds; i.e. 10 minutes
MAX_WAIT_TIME_FOR_PARTIAL_SIGS = 600 # seconds; i.e. 10 minutes


from __future__ import annotations
from base64 import b64decode, b64encode
from context import abstractclasses, musig
from json import dumps, loads
import musig
import nacl.bindings
import unittest


class ExtendedDictExample(abstractclasses.ExtendedDict):
    def __init__(self, data: dict = None):
        if type(data) is dict and 'stuff' in data:
            self.stuff = data['stuff']

    def __bytes__(self) -> bytes:
        if self.stuff is None:
            return b'nothing'
        if type(self.stuff) is str:
            return bytes(self.stuff, 'utf-8')
        return self.stuff if type(self.stuff) is bytes else bytes(self.stuff)

    @classmethod
    def from_bytes(cls, data: bytes) -> ExtendedDictExample:
        if type(data) is not bytes:
            raise TypeError('data must be bytes')

        return cls({
            'stuff': data
        })

    @property
    def stuff(self):
        return self._stuff if hasattr(self, '_stuff') else None

    @stuff.setter
    def stuff(self, data):
                if type(data) not in (str, bytes):
                    raise TypeError('stuff must be either str or bytes')
                self['stuff'] = data


class AbstractNonceExample(abstractclasses.AbstractNonce):
    def __init__(self, data: dict = None):
        if data is None:
            self.r = bytes.fromhex('e7b8484e312a35b8b60a8998458b5a11c7d5b39503be2855b4d54938a34fae67')
            self.R = bytes.fromhex('0a622ae3c831c30dd90e7f7e1477148be3d6725ad402f3809961f0b27d5e2c3f')

        if type(data) is dict and 'r' in data:
            self.r = data['r'] if type(data['r']) is bytes else b64decode(data['r'])

        if type(data) is dict and 'R' in data:
            self.R = data['R'] if type(data['R']) is bytes else b64decode(data['R'])

        if self.R is None and self.r is not None:
            self.R = nacl.bindings.crypto_scalarmult_ed25519_base_noclamp(self.r)

    def __bytes__(self) -> bytes:
        if self.r is not None:
            return b'r' + self.r
        if self.R is not None:
            return self.R
        return b''

    @classmethod
    def from_bytes(cls, data: bytes) -> AbstractNonceExample:
        if type(data) is not bytes:
            return TypeError('data must be bytes of len 32 or 33')
        if len(data) not in (32, 33):
            return ValueError('data must be bytes of len 32 or 33')

        return cls({
            ('R' if len(data) == 32 else 'r'): data[-32:]
        })

    def __add__(self, other) -> AbstractNonceExample:
        if not isinstance(other, AbstractNonceExample):
            raise TypeError('cannot add AbstractNonceExample to non-AbstractNonceExample')
        R_sum = musig.aggregate_points([self.R, other.R])
        return self.__class__.from_bytes(R_sum)

    def copy(self) -> AbstractNonceExample:
        return self.__class__({**self})

    def public(self) -> AbstractNonceExample:
        return self.__class__({'R': self.R})

    # properties
    @property
    def r(self):
        return self._r if hasattr(self, '_r') else None

    @r.setter
    def r(self, value):
        if not isinstance(value, bytes):
            raise TypeError('r value must be bytes')
        if len(value) != 32:
            raise ValueError('r value must have length 32')

        self['r'] = value

    @property
    def R(self):
        return self._R if hasattr(self, '_R') else None

    @R.setter
    def R(self, value):
        if not isinstance(value, bytes):
            raise TypeError('R value must be bytes')
        if len(value) != 32:
            raise ValueError('R value must have length 32')

        self['R'] = value


class TestMuSigAbstractClasses(unittest.TestCase):
    """Test suite for some abstract classes.
        The purpose of this suite is to test the functionality written in the
        abstract classes themselves, not to test that every abstract class
        can be implemented properly. Everything in the module except
        ProtocolState and ProtocolError derives from ExtendedDict, so the
        primary purpose is to test the features built into that class and show
        how to implement derivative classes properly through the examples of
        ExtendedDictExample and AbstractNonceExample.
    """
    def test_class_deriving_from_ExtendedDict_cannot_set_nonproperty_keys(self):
        t = ExtendedDictExample()
        assert t is not None

        # positive case: defined property can be set
        assert hasattr(t, 'stuff')
        assert t.stuff is None
        assert 'stuff' not in t
        t['stuff'] = 'hello world'
        assert t.stuff == 'hello world'
        assert 'stuff' in t

        # negative case: undefined property cannot be set
        assert not hasattr(t, 'notstuff')
        t['notstuff'] = 'will not work'
        assert 'notstuff' not in t

    def test_class_deriving_from_ExtendedDict_sets_properties_and_key_value_pairs_properly(self):
        t = ExtendedDictExample()
        # case one: use the property setter
        t.stuff = b'hello world'
        assert t['stuff'] == b64encode(t.stuff).decode()
        # case two: use the __setitem__ implicit functionality
        t['stuff'] = b'123abc'
        assert t.stuff == b'123abc'
        assert t['stuff'] == b64encode(t.stuff).decode()

    def test_class_deriving_from_ExtendedDict_serializes_and_deserializes_properly(self):
        t0 = ExtendedDictExample({'stuff': '123'})
        t00 = ExtendedDictExample({'stuff': '321'})
        assert t0 != t00

        bts1 = bytes(t0)
        bts11 = bytes(t00)
        assert bts1 != bts11
        str1 = str(t0)
        str11 = str(t00)
        assert str1 != str11

        t1 = ExtendedDictExample.from_bytes(bts1)
        t11 = ExtendedDictExample.from_bytes(bts11)
        assert t1 != t11
        assert t1 == t0
        assert t11 == t00

        t2 = ExtendedDictExample.from_str(str1)
        t22 = ExtendedDictExample.from_str(str11)
        assert t2 != t22
        assert t2 == t0
        assert t22 == t00

    def test_class_deriving_from_ExtendedDict_is_hashable(self):
        t1 = ExtendedDictExample({'stuff':'things'})
        t2 = ExtendedDictExample({'stuff':'not things'})
        t3 = ExtendedDictExample({**t1})

        ts = set([t1, t2, t3])
        assert len(list(ts)) == 2

    def test_class_deriving_from_AbstractNonce_cannot_set_nonproperty_keys(self):
        n = AbstractNonceExample()
        assert 'r' in n and n.r is not None
        assert 'R' in n and n.R is not None
        n['notaproperty'] = 'will not be set'
        assert 'notaproperty' not in n

    def test_class_deriving_from_AbstractNonce_serializes_and_deserializes_properly(self):
        n0 = AbstractNonceExample()
        n00 = n0.public()
        bts0 = bytes(n0)
        bts00 = bytes(n00)
        str0 = str(n0)
        str00 = str(n00)
        js0 = dumps(n0)
        js00 = dumps(n00)

        n1 = AbstractNonceExample.from_bytes(bts0)
        n11 = AbstractNonceExample.from_bytes(bts00)
        n2 = AbstractNonceExample.from_str(str0)
        n22 = AbstractNonceExample.from_str(str00)
        n3 = AbstractNonceExample(loads(js0))
        n33 = AbstractNonceExample(loads(js00))

        assert n0 != n00
        assert n0 == n1
        assert n00 == n11
        assert n0 == n2
        assert n00 == n22
        assert n0 == n3
        assert n00 == n33

        assert n0.r is not None and n0.R is not None
        assert n00.r is None and n00.R is not None
        assert n1.r is not None and n1.R is not None
        assert n11.r is None and n11.R is not None
        assert n2.r is not None and n2.R is not None
        assert n22.r is None and n22.R is not None
        assert n3.r is not None and n3.R is not None
        assert n33.r is None and n33.R is not None


if __name__ == '__main__':
    unittest.main()



"""A collection of helper functions that are used by many classes. Not all
    calls to nacl.bindings functions are abstracted into helpers, but many are.
    Each function has a short docblock explaining its purpose.
"""


from hashlib import new
from nacl.signing import SigningKey, VerifyKey
import nacl.bindings


def clamp_scalar(scalar: bytes|SigningKey, from_private_key: bool = False) -> bytes:
    """Make a clamped scalar."""
    if type(scalar) is bytes and len(scalar) >= 32:
        x_i = bytearray(scalar[:32])
    elif type(scalar) is SigningKey:
        x_i = bytearray(new('sha512', bytes(scalar)).digest()[:32])
        from_private_key = True
    else:
        raise ValueError('not a SigningKey and not 32+ bytes scalar')

    if from_private_key:
        # set bits 0, 1, and 2 to 0
        # nb: lsb is right-indexed
        x_i[0] &= 0b11111000
        # set bit 254 to 1
        x_i[31] |= 0b01000000

    # set bit 255 to 0
    x_i[31] &= 0b01111111

    return bytes(x_i)

def aggregate_points(points: list[bytes|VerifyKey]) -> bytes:
    """Aggregate points on the Ed25519 curve."""
    # type checking inputs
    for pt in points:
        if type(pt) is not bytes and type(pt) is not VerifyKey:
            raise TypeError('each point must be bytes or VerifyKey')

    # normalize points to bytes
    points = [pt if type(pt) is bytes else bytes(pt) for pt in points]

    # raise an error for invalid points
    for pt in points:
        if not nacl.bindings.crypto_core_ed25519_is_valid_point(pt):
            raise ValueError('each point must be a valid ed25519 point')

    # compute the sum
    sum = points[0]
    for i in range(1, len(points)):
        sum = nacl.bindings.crypto_core_ed25519_add(sum, points[i])

    return sum

def H_big(*parts: bytes) -> bytes:
    """The big, 64-byte hash function."""
    return new('sha512', b''.join(parts)).digest()

def H_small(*parts: bytes) -> bytes:
    """The small, 32-byte hash function."""
    return nacl.bindings.crypto_core_ed25519_scalar_reduce(H_big(*parts))

def derive_key_from_seed(seed: bytes) -> bytes:
    """Derive the scalar used for signing from a seed."""
    return clamp_scalar(H_big(seed)[:32], True)

def derive_challenge(L: bytes, X_i: bytes, X: bytes, R: bytes, M: bytes) -> bytes:
    """Derive the challenge used for making a partial signature."""
    a_i = H_agg(L, X_i)
    c = H_sig(R, X, M)
    return clamp_scalar(nacl.bindings.crypto_core_ed25519_scalar_mul(a_i, c))

def H_agg(L: bytes, X_i: bytes) -> bytes:
    """The hash used for aggregating keys."""
    return clamp_scalar(H_small(L, X_i))

def H_sig(R: bytes, X: bytes, M: bytes) -> bytes:
    """The hash used to derive the challenge used in signing and verifying."""
    return clamp_scalar(H_small(R, X, M))

def xor(b1: bytes, b2: bytes) -> bytes:
    """XOR two equal-length byte strings together."""
    b3 = bytearray()
    for i in range(len(b1)):
        b3.append(b1[i] ^ b2[i])

    return bytes(b3)

def bytes_are_same(b1: bytes, b2: bytes) -> bool:
    """Timing-attack safe bytes comparison."""
    return len(b1) == len(b2) and int.from_bytes(xor(b1, b2), 'little') == 0



from context import musig
from json import dumps, loads
from nacl.signing import SigningKey, VerifyKey
import inspect
import musig
import unittest


class TestMuSigSomething(unittest.TestCase):
    """Test suite for Something."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]

    def test_SingleSigKey_is_a_class(self):
        assert hasattr(musig, 'SingleSigKey')
        assert inspect.isclass(musig.SingleSigKey)

    def test_SingleSigKey_init_raises_ValueError_when_given_None_param(self):
        with self.assertRaises(ValueError) as err:
            musig.SingleSigKey()
        assert str(err.exception) == 'cannot instantiate an empty SingleSigKey'
        with self.assertRaises(TypeError) as err:
            musig.SingleSigKey('not a dict')
        assert str(err.exception) == 'data for initialization must be of type dict'

    def test_SingleSigKey_instances_have_correct_attributes(self):
        ssk = musig.SingleSigKey({'skey': self.signing_keys[0]})
        assert hasattr(ssk, 'skey') and isinstance(ssk.skey, SigningKey)
        assert hasattr(ssk, 'vkey') and isinstance(ssk.vkey, musig.PublicKey)
        assert hasattr(ssk, 'vkey_base') and isinstance(ssk.vkey_base, VerifyKey)
        assert ssk.skey == self.signing_keys[0]
        assert ssk.vkey_base == self.signing_keys[0].verify_key

    def test_SingleSigKey_from_bytes_raises_ValueError_or_TypeError_for_invalid_params(self):
        with self.assertRaises(ValueError) as err:
            musig.SingleSigKey.from_bytes(b'invalid bytes')
        assert str(err.exception) == 'bytes input must have length of 32'
        with self.assertRaises(TypeError):
            musig.SingleSigKey.from_bytes('not bytes')
        assert str(err.exception) == 'bytes input must have length of 32'

    def test_SingleSigKey_instances_serialize_and_deserialize_properly(self):
        ssk0 = musig.SingleSigKey({'skey': self.signing_keys[0]})
        bts1 = bytes(ssk0)
        str2 = str(ssk0)
        js = dumps(ssk0)
        ssk1 = musig.SingleSigKey.from_bytes(bts1)
        ssk2 = musig.SingleSigKey.from_str(str2)
        ssk3 = musig.SingleSigKey(loads(js))

        assert type(bts1) is bytes and len(bts1) == 32
        assert type(str2) is str and len(str2) == 64
        assert type(js) is str

        assert ssk0 == ssk1
        assert ssk0 == ssk2
        assert ssk0 == ssk3

    def test_SingleSigKey_sign_message_method_returns_Signature(self):
        ssk = musig.SingleSigKey({'skey': self.signing_keys[0]})
        sig = ssk.sign_message(b'hello world')
        assert isinstance(sig, musig.Signature)

    def test_SingleSigKey_signatures_are_verified_by_relevant_PublicKey(self):
        ssk = musig.SingleSigKey({'skey': self.signing_keys[0]})
        sig = ssk.sign_message(b'hello world')
        pubkey = ssk.vkey.public()
        assert pubkey.verify(sig)


if __name__ == '__main__':
    unittest.main()



from context import musig
from json import dumps, loads
from nacl.signing import SigningKey
import inspect
import unittest


class TestMuSigPartialSignature(unittest.TestCase):
    """Test suite for PartialSignature."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]

    def test_PartialSignature_is_a_class(self):
        assert inspect.isclass(musig.PartialSignature)

    def test_PartialSignature_init_raises_TypeError_when_called_with_invalid_input(self):
        with self.assertRaises(TypeError) as err:
            musig.PartialSignature('not a dict')
        assert str(err.exception) == 'data must be type dict'

    def test_PartialSignature_instances_have_correct_attributes(self):
        pkey = musig.PublicKey.create(self.verify_keys)
        n = musig.Nonce()
        psig = musig.PartialSignature.create(self.signing_keys[0], n.r, pkey.L,
            pkey, n.R, b'hello world')
        assert hasattr(psig, 'c_i') and type(psig.c_i) is bytes
        assert hasattr(psig, 's_i') and type(psig.s_i) is bytes
        assert hasattr(psig, 'R') and type(psig.R) is bytes
        assert hasattr(psig, 'M') and type(psig.M) is bytes

    def test_PartialSignature_from_bytes_raises_ValueError_or_TypeError_when_given_invalid_serialization(self):
        with self.assertRaises(TypeError) as err:
            musig.PartialSignature.from_bytes('not bytes')
        assert str(err.exception) == 'data must be bytes of len == 32 or >=96'
        with self.assertRaises(ValueError) as err:
            musig.PartialSignature.from_bytes(b'invalid bytes')
        assert str(err.exception) == 'data must be bytes of len == 32 or >=96'

    def test_PartialSignature_instances_serialize_and_deserialize_properly(self):
        pkey = musig.PublicKey.create(self.verify_keys[0:1])
        n = musig.Nonce()
        M = b'hello world'
        ps0 = musig.PartialSignature.create(self.signing_keys[0], n.r, pkey.L,
            pkey, n.R, M)
        str1 = str(ps0)
        bts = bytes(ps0)
        js = dumps(ps0)
        ps1 = musig.PartialSignature.from_str(str1)
        ps2 = musig.PartialSignature.from_bytes(bts)
        ps3 = musig.PartialSignature(loads(js))

        assert type(str1) is str and len(str1) == 96*2 + len(M)*2
        assert type(js) is str
        assert ps0 == ps1
        assert ps1 == ps2
        assert ps2 == ps3

    def test_PartialSignature_public_method_returns_instance_with_only_s_i(self):
        pkey = musig.PublicKey.create(self.verify_keys[0:1])
        n = musig.Nonce()
        M = b'hello world'
        ps0 = musig.PartialSignature.create(self.signing_keys[0], n.r, pkey.L,
            pkey, n.R, M)
        ps1 = ps0.public()

        assert ps1.s_i is not None
        assert ps1.s_i == ps0.s_i
        assert ps1.c_i is None
        assert ps1.R is None
        assert ps1.M is None

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from abc import abstractmethod
from base64 import b64encode
from enum import EnumMeta
from musig import bytes_are_same
from nacl.signing import SigningKey, VerifyKey
from uuid import UUID


class ExtendedDict(dict):
    """ExtendedDict handles some serialization/deserialization."""
    def __setitem__(self, key, value) -> None:
        """Method that is called when `instance[key]=value` is used.
            Any key that is not a property of the instance will not be set.
            Any key that is set will have its values encoded to be maximally
            compatible with json serialization.
        """
        if hasattr(self, key):
            setattr(self, f'_{key}', value)
            if type(value) in (int, str):
                super().__setitem__(key, value)
            elif type(value) is bytes:
                super().__setitem__(key, b64encode(value).decode())
            elif type(value) in (tuple, list):
                val = []
                for item in value:
                    if type(item) in (int, str):
                        val.append(item)
                    else:
                        val.append(b64encode(item if type(item) is bytes else bytes(item)).decode())
                super().__setitem__(key, tuple(val))
            elif type(value) is dict:
                nv = {}
                for name in value:
                    val = value[name]
                    if type(name) not in (str, int):
                        name = b64encode(name if type(name) is bytes else bytes(name)).decode()
                    if type(val) not in (str, int):
                        val = b64encode(val if type(val) is bytes else bytes(val)).decode()
                    nv[name] = val
                super().__setitem__(key, nv)
            else:
                super().__setitem__(key, b64encode(bytes(value)).decode())

    @abstractmethod
    def __bytes__(self) -> bytes:
        """Result of calling bytes() on instance; i.e. serializes instance as
            bytes. Must be implemented for __str__ and from_str to function.
        """
        ...

    def __str__(self) -> str:
        """Result of calling str() on instance; serializes instance as a
            hexidecimal str based upon the bytes serialization. Relies upon
            __bytes__ to function.
        """
        return bytes(self).hex()

    def __hash__(self) -> int:
        """Result of calling hash() on instance; allows inclusion in sets and
            use as a key in a dict (of dubious worth, but technically possible.)
        """
        return hash(bytes(self))

    def __eq__(self, other) -> bool:
        """Timing-attack safe comparison."""
        if not isinstance(other, self.__class__):
            return False
        return bytes_are_same(bytes(self), bytes(other))

    @classmethod
    @abstractmethod
    def from_bytes(cls, data: bytes) -> ExtendedDict:
        """Deserializes output from __bytes__. Must be implemented for from_str
            to function.
        """
        ...

    @classmethod
    def from_str(cls, data: str) -> ExtendedDict:
        """Deserializes output from __str__, i.e. hexidecimal str of __bytes__
            output. Relies upon from_bytes to function.
        """
        return cls.from_bytes(bytes.fromhex(data))


class AbstractNonce(ExtendedDict):
    @abstractmethod
    def __add__(self, other: AbstractNonce) -> AbstractNonce:
        ...

    @abstractmethod
    def copy(self) -> AbstractNonce:
        """Make a copy of the instance. Heavily dependent upon the __init__ definition."""
        ...

    @abstractmethod
    def public(self) -> AbstractNonce:
        """Return a copy of the instance with only the public point value (R)."""
        ...

    @property
    @abstractmethod
    def r(self) -> bytes:
        """The private scalar value."""
        ...

    @property
    @abstractmethod
    def R(self) -> bytes:
        """The public point value."""
        ...


class AbstractNonceCommitment(ExtendedDict):
    @classmethod
    @abstractmethod
    def create(cls, nonce: AbstractNonce) -> AbstractNonceCommitment:
        """Create a new instance by hashing the given nonce."""
        ...

    @abstractmethod
    def copy(self) -> AbstractNonceCommitment:
        """Make a copy of the instance. Heavily dependent upon the __init__ definition."""
        ...

    @abstractmethod
    def is_valid_for(self, nonce: AbstractNonce) -> bool:
        """Check if the current instance is a valid commitment for a given nonce."""
        ...

    @property
    def HR(self) -> bytes:
        """The nonce commitment bytes (hash of a nonce point)."""
        ...


class AbstractPartialSignature(ExtendedDict):
    @classmethod
    def create(cls, skey: SigningKey, r: bytes, L: bytes, X: AbstractPublicKey,
            R: bytes, M: bytes) -> dict:
        """Create a new instance using the SigningKey of the participant (skey),
            from which the private key will be derived (bytes(skey) returns the
            seed); the private nonce of the participant (r); the keyset encoding
            of the participants (L); the aggregate public key (X); the aggregate
            public nonce point (R); and the message (M).
        """
        ...

    @abstractmethod
    def public(self) -> AbstractPartialSignature:
        """Return a copy of the instance with only the public value (s_i)."""
        ...

    @property
    @abstractmethod
    def c_i(self) -> bytes:
        """The non-interactive challenge for the participant."""
        ...

    @property
    @abstractmethod
    def s_i(self) -> bytes:
        """The partial signature scalar."""
        ...

    @property
    @abstractmethod
    def R(self) -> bytes:
        """The aggregate nonce point."""
        ...

    @property
    @abstractmethod
    def M(self) -> bytes:
        """The message to be signed."""
        ...


class AbstractPublicKey(ExtendedDict):
    @classmethod
    @abstractmethod
    def create(cls, keys: list) -> AbstractPublicKey:
        """Create a new instance from a list/tuple of participant VerifyKeys.
            This must derive the keyset encoding (L) and the aggregate key
            (denoted X in the MuSig paper and gvkey in this class).
        """
        ...

    @abstractmethod
    def public(self) -> AbstractPublicKey:
        """Return a copy of the instance with only the public value (gvkey)."""
        ...

    @abstractmethod
    def verify(self, sig: AbstractSignature) -> bool:
        """Check if a given signature is valid for the current aggregate key."""
        ...

    @classmethod
    @abstractmethod
    def aggregate_public_keys(cls, vkeys: list[VerifyKey|bytes], key_set_L: bytes = None) -> bytes:
        """Calculate the aggregate public key from the participant keys."""
        ...

    @classmethod
    @abstractmethod
    def pre_agg_key_transform(cls, vkey: VerifyKey, key_set_L: bytes) -> bytes:
        """Transform a participant VerifyKey prior to calculating the aggregate
            public key. This is called by aggregate_public_keys on every
            participant VerifyKey.
        """
        ...

    @classmethod
    @abstractmethod
    def encode_key_set(cls, vkeys: list) -> bytes:
        """Sort the participant keys into a deterministic order, then hash the
            list to produce the keyset encoding (L).
        """
        ...

    @property
    @abstractmethod
    def L(self) -> bytes:
        """The keyset encoding used for calculating partial signatures."""
        ...

    @property
    @abstractmethod
    def gvkey(self) -> bytes:
        """The bytes of the aggregate key (denoted X in the MuSig paper)."""
        ...

    @property
    @abstractmethod
    def vkeys(self) -> tuple[VerifyKey, ...]:
        """Tuple of untransformed participant VerifyKeys."""
        ...


class AbstractSignature(ExtendedDict):
    @classmethod
    @abstractmethod
    def create(cls, R: bytes, M: bytes, parts: list[AbstractPartialSignature]) -> AbstractSignature:
        """Create a new instance using the aggregate nonce point (R), the
            message (M), and the list/tuple of partial signatures (scalars s_i).
        """
        ...

    @property
    @abstractmethod
    def R(self) -> bytes:
        """Aggregate nonce point."""
        ...

    @property
    @abstractmethod
    def s(self) -> bytes:
        """Aggregate signature made from summing partial signatures."""
        ...

    @property
    @abstractmethod
    def M(self) -> bytes:
        """Message to be signed."""
        ...

    @property
    @abstractmethod
    def parts(self) -> tuple[AbstractPartialSignature, ...]:
        """Tuple of partial signatures summed together to create the signature."""
        ...


class AbstractSingleSigKey(ExtendedDict):
    @abstractmethod
    def sign_message(self, M: bytes) -> AbstractSignature:
        """Sign a message (M) and return a signature."""
        ...

    @property
    @abstractmethod
    def skey(self) -> SigningKey|None:
        """The SigningKey used for creating signatures."""
        ...

    @property
    @abstractmethod
    def vkey(self) -> AbstractPublicKey|None:
        """The aggregate public key for verifying signatures."""
        ...

    @property
    @abstractmethod
    def vkey_base(self) -> VerifyKey|None:
        """The VerifyKey base used to calculate the aggregate public key."""
        ...


class AbstractProtocolState(EnumMeta):
    """Enum mapping human-readable protocol states to 1-byte ints."""
    ...


class AbstractProtocolMessage(ExtendedDict):
    @abstractmethod
    def parse_message(self) -> None:
        """Parses the message into parts based upon protocol state, e.g. lists
            of participant keys, nonces, nonce commitments, etc, storing the
            result in self.message_parts/self['message_parts'].
        """
        ...

    @abstractmethod
    def add_signature(self, skey: SigningKey) -> AbstractProtocolMessage:
        """Add a signature to the message using the provided SigningKey. This
            signs the protocol state + message, attaching the signature and the
            relevant VerifyKey to the instance.
        """
        ...

    @abstractmethod
    def check_signature(self) -> bool:
        """Return true if and only if the instance contains a VerifyKey and a
            signature valid for the protocol state + message, else return False.
        """
        ...

    @classmethod
    @abstractmethod
    def create(cls, id: UUID|None, state: AbstractProtocolState, data: list) -> AbstractProtocolMessage:
        """Create a new instance with the given id, state, and data.
            NB: id=None should be accepted only for state=EMPTY.
        """
        ...

    @property
    @abstractmethod
    def session_id(self) -> UUID:
        """The UUID of the signing session for which this message was constructed."""
        ...

    @property
    @abstractmethod
    def state(self) -> AbstractProtocolState:
        """The protocol state of the message."""
        ...

    @property
    @abstractmethod
    def message(self) -> bytes:
        """The message itself."""
        ...

    @property
    @abstractmethod
    def message_parts(self) -> list:
        """The things that serialize into and deserialize from self.message."""
        ...

    @property
    @abstractmethod
    def signature(self) -> AbstractSignature:
        """The Signature result of add_signature."""
        ...

    @property
    @abstractmethod
    def vkey(self) -> AbstractPublicKey:
        """The VerifyKey used to verify the signature."""
        ...


class AbstractSigningSession(ExtendedDict):
    @abstractmethod
    def add_participant_keys(self, keys: VerifyKey|list[VerifyKey]) -> None:
        """Add participant VerifyKey(s)."""
        ...

    @abstractmethod
    def add_nonce_commitment(self, commitment: AbstractNonceCommitment, vkey: VerifyKey) -> None:
        """Add a NonceCommitment from a participant identified by the VerifyKey."""
        ...

    @abstractmethod
    def add_nonce(self, nonce: AbstractNonce, vkey: VerifyKey) -> None:
        """Add a Nonce from a participant identified by the VerifyKey."""
        ...

    @abstractmethod
    def make_partial_signature(self) -> AbstractPartialSignature:
        """Create a partial signature to be broadcast to other participants."""
        ...

    @abstractmethod
    def add_partial_signature(self, sig: AbstractPartialSignature, vkey: VerifyKey) -> None:
        """Add a PartialSignature from a participant identified by the VerifyKey."""
        ...

    @abstractmethod
    def update_protocol_state(self) -> None:
        """Handle transitions between ProtocolStates as the SigningSession values
            are updated.
        """
        ...

    @property
    @abstractmethod
    def id(self) -> UUID:
        """The UUID of the session."""
        ...

    @property
    @abstractmethod
    def number_of_participants(self) -> int:
        """The number of participants expected to participate in the protocol."""
        ...

    @property
    @abstractmethod
    def protocol_state(self) -> AbstractProtocolState:
        """The current state of the session."""
        ...

    @property
    @abstractmethod
    def last_updated(self) -> int:
        """A timestamp recording the last time the protocol state was updated."""
        ...

    @property
    @abstractmethod
    def skey(self) -> SigningKey:
        """The SigningKey of the participant using this instance."""
        ...

    @property
    @abstractmethod
    def vkeys(self) -> tuple[VerifyKey, ...]:
        """A tuple of participant VerifyKeys."""
        ...

    @property
    @abstractmethod
    def nonce_commitments(self) -> dict[VerifyKey, AbstractNonceCommitment]:
        """A dict mapping participant VerifyKey to NonceCommitment."""
        ...

    @property
    @abstractmethod
    def nonce_points(self) -> dict[VerifyKey, AbstractNonce]:
        """A dict mapping participant VerifyKey to Nonce. Note that the Nonce
            for the participant using this instance will include the private
            scalar value, but the Nonces of other participants will include only
            the public point values.
        """
        ...

    @property
    @abstractmethod
    def aggregate_nonce(self) -> AbstractNonce:
        """The aggregate nonce point for the session."""
        ...

    @property
    @abstractmethod
    def message(self) -> bytes:
        """The message to be n-of-n signed."""
        ...

    @property
    @abstractmethod
    def partial_signatures(self) -> dict[VerifyKey, AbstractPartialSignature]:
        """A dict mapping participant VerifyKey to PartialSignature (public
            values s_i only).
        """
        ...

    @property
    @abstractmethod
    def public_key(self) -> AbstractPublicKey:
        """The aggregate public key for the session."""
        ...

    @property
    @abstractmethod
    def signature(self) -> AbstractSignature:
        """The final n-of-n signature."""
        ...



from __future__ import annotations
from base64 import b64decode
from musig.abstractclasses import AbstractSignature, AbstractPartialSignature
from musig.partialsignature import PartialSignature
import nacl.bindings


class Signature(AbstractSignature):
    """A class that sums PartialSignatures into a full Signature."""
    def __init__(self, data: dict = None) -> None:
        """Initialize an instance. Initialize with `{'parts': list_of_partial_sigs,
            'M': bytes, 'R': bytes}` to create a new signature from parts. Initialize
            with `{'s': bytes, 'R': bytes, 'M': bytes}` to restore a signature.
        """
        if not isinstance(data, dict):
            raise TypeError('data for initialization must be of type dict')

        if 'R' in data:
            self.R = data['R'] if type(data['R']) is bytes else b64decode(data['R'])
        if 'M' in data:
            self.M = data['M'] if type(data['M']) is bytes else b64decode(data['M'])
        if 's' in data:
            self.s = data['s'] if type(data['s']) is bytes else b64decode(data['s'])
        if 'parts' in data:
            parts = [PartialSignature.from_bytes(p if type(p) is bytes else b64decode(p)) for p in data['parts']]
            self.parts = parts

        if len(self.parts) > 0 and self.R is not None and self.M is not None and self.s is None:
            sig = self.create(self.R, self.M, self.parts)
            self.s = sig.s

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on an instance; i.e. serialize to bytes."""
        return self.R + self.s + self.M

    @classmethod
    def from_bytes(cls, data: bytes) -> Signature:
        """Deserializes output from __bytes__."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes with length at least 65')
        if len(data) < 65:
            raise ValueError('data must be bytes with length at least 65')

        R = data[:32]
        s = data[32:64]
        M = data[64:]

        return cls({
            'R': R,
            's': s,
            'M': M
        })

    @classmethod
    def create(cls, R: bytes, M: bytes, parts: list[AbstractPartialSignature]) -> Signature:
        """Create a new instance using the aggregate nonce point (R), the
            message (M), and the list/tuple of partial signatures (scalars s_i).
        """
        if type(R) is not bytes or \
                len(R) != nacl.bindings.crypto_core_ed25519_BYTES or \
                not nacl.bindings.crypto_core_ed25519_is_valid_point(R):
            raise ValueError('R (aggregate nonce point) must be a valid ed25519 point')

        if type(parts) not in (list, tuple) or len(parts) < 1:
            raise ValueError('parts must be list or tuple of PartialSignature objects')

        for p in parts:
            if not isinstance(p, AbstractPartialSignature):
                raise ValueError('parts must be list or tuple of PartialSignature objects')

        # sum the partial signatures
        s = parts[0].s_i
        for i in range(1, len(parts)):
            s = nacl.bindings.crypto_core_ed25519_scalar_add(s, parts[i].s_i)

        return cls({
            'R': R,
            's': s,
            'M': M,
            'parts': [bytes(p) for p in parts]
        })

    @property
    def R(self) -> bytes:
        """Aggregate nonce point."""
        return self._R if hasattr(self, '_R') else None

    @R.setter
    def R(self, data: bytes):
        """Aggregate nonce point."""
        if type(data) is not bytes:
            raise TypeError('R must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('R must be bytes of len 32')

        self['R'] = data

    @property
    def s(self) -> bytes:
        """Aggregate signature made from summing partial signatures."""
        return self._s if hasattr(self, '_s') else None

    @s.setter
    def s(self, data: bytes):
        """Aggregate signature made from summing partial signatures."""
        if type(data) is not bytes:
            raise TypeError('s must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('s must be bytes of len 32')

        self['s'] = data

    @property
    def M(self) -> bytes:
        """Message to be signed."""
        return self._M if hasattr(self, '_M') else None

    @M.setter
    def M(self, data: bytes):
        """Message to be signed."""
        if type(data) not in (bytes, str):
            raise TypeError('M must be bytes or str')

        self['M'] = data if type(data) is bytes else bytes(data, 'utf-8')

    @property
    def parts(self) -> tuple[AbstractPartialSignature, ...]:
        """Tuple of partial signatures summed together to create the signature."""
        return self._parts if hasattr(self, '_parts') else tuple()

    @parts.setter
    def parts(self, data: list[AbstractPartialSignature]):
        """Tuple of partial signatures summed together to create the signature."""
        if type(data) not in (list, tuple):
            raise TypeError('parts must be list or tuple of PartialSignatures')
        for ps in data:
            if not isinstance(ps, AbstractPartialSignature):
                raise TypeError('parts must be list or tuple of PartialSignatures')

        self['parts'] = tuple(data)



from __future__ import annotations
from base64 import b64decode
from musig.abstractclasses import AbstractPublicKey, AbstractPartialSignature
from musig.helpers import derive_key_from_seed, derive_challenge
from nacl.signing import SigningKey
import nacl.bindings


class PartialSignature(AbstractPartialSignature):
    """A class that handles creation, serialization, and deserialization of
        partial signatures used to create MuSig aggregate signatures.
    """
    def __init__(self, data: dict) -> None:
        """Initialize an instance with dict.
            At a minimum data should include s_i key with base64-encoded value
            of 32 bytes; keys of c_i and R with base64-encoded value of 32 bytes
            each are optional; key of M with base64-encoded value of bytes is
            also optional.
        """
        if not isinstance(data, dict):
            raise TypeError('data must be type dict')

        if 'c_i' in data:
            self.c_i = data['c_i'] if type(data['c_i']) is bytes else b64decode(data['c_i'])
        if 's_i' in data:
            self.s_i = data['s_i'] if type(data['s_i']) is bytes else b64decode(data['s_i'])
        if 'R' in data:
            self.R = data['R'] if type(data['R']) is bytes else b64decode(data['R'])
        if 'M' in data:
            self.M = data['M'] if type(data['M']) is bytes else b64decode(data['M'])

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on an instance."""
        if self.c_i is not None and self.R is not None and self.M is not None:
            return self.s_i + self.c_i + self.R + self.M
        return self.s_i

    @classmethod
    def from_bytes(cls, data: bytes) -> PartialSignature:
        """Deserializes output from __bytes__."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes of len == 32 or >=96')

        if len(data) < 32 or (len(data) > 32 and len(data) < 96):
            raise ValueError('data must be bytes of len == 32 or >=96')

        new_data = {
            's_i': data[:32]
        }

        if len(data) >= 96:
            new_data['c_i'] = data[32:64]
            new_data['R'] = data[64:96]

        if len(data) > 96:
            new_data['M'] = data[96:]

        return cls(new_data)

    @classmethod
    def create(cls, skey: SigningKey, r_i: bytes, L: bytes, X: AbstractPublicKey,
            R: bytes, M: bytes) -> dict:
        """Create a new instance using the SigningKey of the participant (skey),
            from which the private key will be derived (bytes(skey) returns the
            seed); the private nonce of the participant (r_i); the keyset encoding
            of the participants (L); the aggregate public key (X); the aggregate
            public nonce point (R); and the message (M).
        """
        x_i = derive_key_from_seed(bytes(skey))
        c_i = derive_challenge(L, bytes(skey.verify_key), bytes(X.public()), R, M)
        s_i = nacl.bindings.crypto_core_ed25519_scalar_mul(c_i, x_i)
        s_i = nacl.bindings.crypto_core_ed25519_scalar_add(r_i, s_i)

        return cls({
            'c_i': c_i,
            's_i': s_i,
            'R': R,
            'M': M,
        })

    def public(self) -> PartialSignature:
        """Return a copy of the instance with only the public value (s_i)."""
        return self.__class__({'s_i': self.s_i})

    @property
    def c_i(self) -> bytes|None:
        """The non-interactive challenge for the participant."""
        return self._c_i if hasattr(self, '_c_i') else None

    @c_i.setter
    def c_i(self, data: bytes):
        """The non-interactive challenge for the participant."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('data must be bytes of len 32')

        self['c_i'] = data

    @property
    def s_i(self) -> bytes|None:
        """The partial signature scalar."""
        return self._s_i if hasattr(self, '_s_i') else None

    @s_i.setter
    def s_i(self, data: bytes):
        """The partial signature scalar."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('data must be bytes of len 32')

        self['s_i'] = data

    @property
    def R(self) -> bytes|None:
        """The aggregate nonce point."""
        return self._R if hasattr(self, '_R') else None

    @R.setter
    def R(self, data: bytes):
        """The aggregate nonce point."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('data must be bytes of len 32')

        self['R'] = data

    @property
    def M(self) -> bytes|None:
        """The message to be signed."""
        return self._M if hasattr(self, '_M') else None

    @M.setter
    def M(self, data: bytes):
        """The message to be signed."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes')

        self['M'] = data



from context import musig
from nacl.signing import SigningKey, VerifyKey, SignedMessage
from secrets import token_bytes
import unittest


class TestMuSigE2E(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.seeds = [token_bytes(32) for _ in range(3)]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]

    def test_1_of_1_musig(self):
        seed = token_bytes(32)
        skey = SigningKey(seed)
        ssk = musig.SingleSigKey({'skey': skey})
        message = b'hello world'
        sig = ssk.sign_message(message)
        pubkey = ssk.vkey
        assert pubkey.verify(sig)

        vkey = VerifyKey(bytes(pubkey.public()))
        assert vkey.verify(SignedMessage(bytes(sig))) == message

    def test_2_of_2_musig(self):
        # simulate sending a message
        def send(pm: musig.ProtocolMessage):
            ...
        # simulate receiving a message
        def receive(pm: musig.ProtocolMessage):
            return musig.ProtocolMessage.from_bytes(bytes(pm))

        # initialize the session
        session = musig.SigningSession({
            'skey': self.signing_keys[0],
            'number_of_participants': 2
        })

        # create and add the values needed for this participant
        n = musig.Nonce()
        nc = musig.NonceCommitment.create(n)
        session.add_nonce_commitment(nc, self.verify_keys[0])
        session.add_nonce(n, self.verify_keys[0])

        # simulate the other guy
        other_skey = self.signing_keys[1]
        pkey = musig.PublicKey.create([self.verify_keys[0], other_skey.verify_key])
        other_n = musig.Nonce()
        other_nc = musig.NonceCommitment.create(other_n)
        agg_n = n + other_n
        other_ps = musig.PartialSignature.create(other_skey, other_n.r, pkey.L, pkey.public(),
            agg_n.R, b'obviously a bitcoin transaction')

        # at this point, the state will be AWAITING_PARTICIPANT_KEYS
        # this can be communicated via use of the ProtocolMessage class, which will
        # be demonstrated just for this step. Assume there are some methods, send
        # and receive, that handle message exchange for us. I am excluding the
        # receiving and parsing of ACK messages (and related resending of un-ACKed
        # messages) for brevity.
        pm = musig.ProtocolMessage.create(session.id, session.protocol_state, [])
        pm = musig.ProtocolMessage.create(
            session.id, musig.ProtocolState.SENDING_PARTICIPANT_KEY, [other_skey.verify_key]
        )
        pm = receive(pm.add_signature(other_skey))

        # parse the message and add the key if possible
        try:
            if not pm.check_signature():
                raise Exception('broken signature')

            pm.parse_message()
            if pm.state is musig.ProtocolState.SENDING_PARTICIPANT_KEY:
                vkey = pm.message_parts[0]
                session.add_participant_keys([vkey])
                # Acknowledge receipt of the key (optional).
                pm = musig.ProtocolMessage.create(
                    session.id, musig.ProtocolState.ACK_PARTICIPANT_KEY, [vkey]
                )
                send(pm.add_signature(session.skey))
        except musig.ProtocolError as err:
            # Abort on a ProtocolError, communicating the error to peers first.
            pm = musig.ProtocolMessage.create(session.id, session.protocol_state, [err])
            send(pm.add_signature(session.skey))
            raise err
        except:
            # Retry or something here.
            pass

        # State is now AWAITING_COMMITMENT, so send this participant's commitment.
        pm = musig.ProtocolMessage.create(session.id, musig.ProtocolState.SENDING_COMMITMENT, [nc])
        send(pm.add_signature(session.skey))

        # Get the vkeys for missing nonce commitments and request those commitments.
        vkeys = [vk for vk in session.vkeys if vk not in session.nonce_commitments.keys()]
        pm = musig.ProtocolMessage.create(session.id, session.protocol_state, vkeys)
        send(pm.add_signature(session.skey))

        # Receive the commitment.
        pm = receive(musig.ProtocolMessage.create(
                session.id, musig.ProtocolState.SENDING_COMMITMENT, [other_nc]
            ).add_signature(other_skey))
        try:
            pm.parse_message()
            if pm.state is musig.ProtocolState.SENDING_COMMITMENT:
                nc = pm.message_parts[0]
                session.add_nonce_commitment(nc, pm.vkey)
                pm = musig.ProtocolMessage.create(session.id, musig.ProtocolState.ACK_COMMITMENT, [nc])
                send(pm.add_signature(session.skey))
        except musig.ProtocolError as err:
            # Abort on a ProtocolError, communicating the error to peers first.
            pm = musig.ProtocolMessage.create(session.id, session.protocol_state, [err])
            send(pm.add_signature(session.skey))
            raise err
        except:
            # Retry or something here.
            pass

        # At this point, I will exclude the message-passing simulation for brevity.

        # State is now AWAITING_MESSAGE. Once the message is agreed upon, we advance.
        session.message = b'obviously a bitcoin transaction'

        # State is now AWAITING_NONCE. Get the nonce point from the other participant.
        # Once all nonce points are gathered, the aggregate nonce will be calculated.
        try:
            session.add_nonce(other_n, other_skey.verify_key)
        except musig.ProtocolError:
            # If the nonce does not match the commitment, if the vkey is not from a
            # known participant, or if too much time has passed while waiting, an
            # error will be raised, so deal with that here.
            pass

        # State is now AWAITING_PARTIAL_SIGNATURE, so send ours and get theirs.
        # Once all partial signatures are gathered, the aggregate signature will
        # be calculated.
        ps = session.make_partial_signature()
        session.add_partial_signature(ps, session.skey.verify_key)
        # send ps, receive received_ps and received_vkey
        try:
            session.add_partial_signature(other_ps, other_skey.verify_key)
        except musig.ProtocolError as err:
            # If the received_vkey is not a known participant vkey, or if a
            # conflicting partial signature is added when one already exists, an
            # error will be raised, so handle the error here.
            pass

        # State should now be COMPLETE, so distribute the signature.
        signature = session.signature
        assert session.public_key.verify(signature)

        # compatible with ordinary Ed25519 signature verification
        sig = SignedMessage(bytes(signature))
        vkey = VerifyKey(bytes(session.public_key.public()))
        assert vkey.verify(sig) == session.message

    def test_3_of_3_musig(self):
        session = musig.SigningSession({
            'skey': self.signing_keys[0],
            'number_of_participants': 3
        })

        session.add_participant_keys(self.verify_keys)

        # create the aggregate public key
        pkey = musig.PublicKey.create(self.verify_keys)

        # simulate the other two participants
        nonces = [musig.Nonce() for _ in range(3)]
        nonce_commitments = [musig.NonceCommitment.create(n) for n in nonces]
        for i in range(3):
            session.add_nonce_commitment(nonce_commitments[i], self.verify_keys[i])
            session.add_nonce(nonces[i], self.verify_keys[i])

        # create partial signatures
        session.message = b'super important thing that must be authorized by a quorum'
        partial_signatures = [
            musig.PartialSignature.create(
                self.signing_keys[i],
                nonces[i].r,
                pkey.L,
                pkey.public(),
                session.aggregate_nonce.R,
                session.message
            )
            for i in range(3)
        ]
        for i in range(3):
            session.add_partial_signature(partial_signatures[i], self.verify_keys[i])

        # check the signature
        signature = session.signature
        assert session.public_key.verify(signature)


if __name__ == '__main__':
    unittest.main()



from context import musig
from json import dumps, loads
from nacl.signing import SigningKey
import inspect
import unittest


class TestMuSigSomething(unittest.TestCase):
    """Test suite for Something."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]
        cls.gvkey = 'eecb56e70d2405a849aa5e55b6e2f96aac2957dba72f8c289994c842e33ec477'

    def test_Signature_is_a_class(self):
        assert inspect.isclass(musig.Signature)

    def test_Signature_init_raises_TypeError_when_called_without_proper_params(self):
        with self.assertRaises(TypeError) as err:
            musig.Signature()
        assert str(err.exception) == 'data for initialization must be of type dict'
        with self.assertRaises(TypeError):
            musig.Signature('not a dict')
        assert str(err.exception) == 'data for initialization must be of type dict'

    def test_Signature_instances_have_correct_attributes(self):
        pkey = musig.PublicKey.create([self.signing_keys[0].verify_key])
        n1 = musig.Nonce()
        M = b'hello world'
        ps1 = musig.PartialSignature.create(self.signing_keys[0], n1.r, pkey.L,
            pkey, n1.R, M)
        sig = musig.Signature.create(n1.R, M, [ps1])
        assert hasattr(sig, 'R') and type(sig.R) is bytes
        assert hasattr(sig, 'M') and type(sig.M) is bytes
        assert hasattr(sig, 's') and type(sig.s) is bytes
        assert hasattr(sig, 'parts') and type(sig.parts) is tuple

    def test_Signature_from_bytes_raises_ValueError_or_TypeError_when_called_with_invalid_serialization(self):
        with self.assertRaises(TypeError) as err:
            musig.Signature.from_bytes('not bytes')
        assert str(err.exception) == 'data must be bytes with length at least 65'
        with self.assertRaises(ValueError) as err:
            musig.Signature.from_bytes(b'invalid bytes')
        assert str(err.exception) == 'data must be bytes with length at least 65'

    def test_Signature_instances_serialize_and_deserialize_properly(self):
        pkey = musig.PublicKey.create([self.signing_keys[0].verify_key])
        n1 = musig.Nonce()
        M = b'hello world'
        ps1 = musig.PartialSignature.create(self.signing_keys[0], n1.r, pkey.L,
            pkey, n1.R, M)
        sig0 = musig.Signature.create(n1.R, M, [ps1])
        bts1 = bytes(sig0)
        str2 = str(sig0)
        js = dumps(sig0)
        sig1 = musig.Signature.from_bytes(bts1)
        sig2 = musig.Signature.from_str(str2)
        sig3 = musig.Signature(loads(js))

        assert type(bts1) is bytes and len(bts1) == 64 + len(M)
        assert type(str2) is str and len(str2) == 2*len(bts1)
        assert type(js) is str

        assert sig1 == sig2
        assert sig1 == sig3

    def test_Signature_create_raises_ValueError_when_supplied_invalid_params(self):
        M = b'hello world'
        n = musig.Nonce()
        pkey = musig.PublicKey.create([s.verify_key for s in self.signing_keys])
        ps = musig.PartialSignature.create(self.signing_keys[0], n.r, pkey.L,
            pkey, n.R, M)
        with self.assertRaises(ValueError):
            musig.Signature.create(b'invalid R bytes', M, [ps])
        with self.assertRaises(ValueError):
            musig.Signature.create(n.R, M, [])
        with self.assertRaises(ValueError):
            musig.Signature.create(n.R, M, [b'sdsd'])


if __name__ == '__main__':
    unittest.main()



from context import musig
from json import dumps, loads
import inspect
import nacl.bindings
import unittest


class TestMuSigNonce(unittest.TestCase):
    """Test suite for Nonce."""
    def test_Nonce_is_a_class(self):
        assert inspect.isclass(musig.Nonce)

    def test_Nonce_instance_has_r_and_R_when_initialized_with_empty_params(self):
        nonce = musig.Nonce()
        assert hasattr(nonce, 'r')
        assert type(nonce.r) is bytes
        assert len(nonce.r) == nacl.bindings.crypto_core_ed25519_SCALARBYTES
        assert hasattr(nonce, 'R')
        assert type(nonce.R) is bytes
        assert nacl.bindings.crypto_core_ed25519_is_valid_point(nonce.R)
        assert 'r' in nonce
        assert type(nonce['r']) is str
        assert 'R' in nonce
        assert type(nonce['R']) is str

    def test_Nonce_instance_cannot_set_key_other_than_r_and_R(self):
        nonce = musig.Nonce()
        nonce['arbitrary_key'] = 'arbitrary value'
        assert 'arbitrary_key' not in nonce
        assert not hasattr(nonce, 'arbitrary_key')

    def test_Nonce_instances_add_together_and_return_point(self):
        n1 = musig.Nonce()
        n2 = musig.Nonce()
        n3 = n1 + n2
        assert n3 != n1
        assert n3 != n2
        assert isinstance(n3, musig.Nonce)
        assert hasattr(n3, 'R')
        assert nacl.bindings.crypto_core_ed25519_is_valid_point(n3.R)
        assert n3.r == None

    def test_Nonce_instance_public_method_returns_Nonce_without_private_value(self):
        n0 = musig.Nonce()
        assert hasattr(n0, 'public')
        assert inspect.ismethod(n0.public)
        n1 = n0.public()
        assert isinstance(n1, musig.Nonce)
        assert n1.r is None
        assert type(n1.R) is bytes

    def test_Nonce_from_bytes_raises_TypeError_or_ValueError_when_given_invalid_data(self):
        with self.assertRaises(TypeError) as err:
            musig.Nonce.from_bytes('not bytes')
        assert str(err.exception) == 'cannot call from_bytes with non-bytes param'
        with self.assertRaises(ValueError) as err:
            musig.Nonce.from_bytes(b'not len 32')
        assert str(err.exception) == 'byte length must be 32 or 33'

    def test_Nonce_instances_serialize_and_deserialize_properly(self):
        n0 = musig.Nonce()
        bts = bytes(n0)
        str1 = str(n0)
        js = dumps(n0)
        n1 = musig.Nonce.from_bytes(bts)
        n2 = musig.Nonce.from_str(str1)
        n3 = musig.Nonce(loads(js))

        assert type(bts) is bytes
        assert type(str1) is str
        assert type(js) is str
        assert n0.r == n1.r and n0.R == n1.R
        assert n0.r == n2.r and n0.R == n2.R
        assert n0.r == n3.r and n0.R == n3.R

    def test_Nonce_instances_with_only_public_value_serialize_and_deserialize_properly(self):
        n0 = musig.Nonce().public()
        bts1 = bytes(n0)
        str1 = str(n0)
        js0 = dumps(n0)
        n1 = musig.Nonce.from_bytes(bts1)
        n2 = musig.Nonce.from_str(str1)
        n3 = musig.Nonce(loads(js0))
        assert n1.r is None and n0.R == n1.R
        assert n2.r is None and n0.R == n2.R
        assert n3.r is None and n0.R == n3.R

    def test_Nonce_instances_can_be_members_of_sets(self):
        n0 = musig.Nonce()
        n1 = n0.copy()
        n2 = musig.Nonce()
        ns = list(set([n0, n1, n2]))
        assert n0 == n1
        assert hash(n0) == hash(n1)
        assert len(ns) == 2


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from base64 import b64decode
from musig.abstractclasses import AbstractPublicKey, AbstractSignature
from musig.helpers import aggregate_points, bytes_are_same, H_agg, H_sig, H_small
from nacl.signing import VerifyKey
import nacl.bindings


class PublicKey(AbstractPublicKey):
    """A class that aggregates the public keys of participants of a Session
        and verifies Signatures.
    """

    def __init__(self, data: dict = None) -> None:
        """Initialize the instance using the given data.
            Initialize with `{'vkeys': list_of_vkeys}` to create a PublicKey from
            participant keys. Initialize with `{'gvkey': bytes}` to restore just
            enough to verify signatures.
        """
        if data is None:
            raise ValueError('cannot instantiate empty PublicKey')

        if type(data) is not dict:
            raise TypeError('input data must be dict')

        if 'vkeys' in data:
            self.vkeys = tuple([vk if isinstance(vk, VerifyKey) else VerifyKey(b64decode(vk)) for vk in data['vkeys']])
        if 'gvkey' in data:
            self.gvkey = data['gvkey'] if type(data['gvkey']) is bytes else b64decode(data['gvkey'])
        if 'L' in data:
            self.L = data['L'] if type(data['L']) is bytes else b64decode(data['L'])

        if len(self.vkeys) > 0 and self.L is None:
            self.L = self.encode_key_set(self.vkeys)

        if len(self.vkeys) > 0 and self.gvkey is None:
            self.gvkey = self.aggregate_public_keys(self.vkeys, self.L)

    def __bytes__(self) -> bytes:
        """Serialize to bytes by calling bytes() on an instance."""
        if len(self.vkeys) > 0:
            return b''.join([bytes(vk) for vk in self.vkeys])
        return self.gvkey

    @classmethod
    def from_bytes(cls, data: bytes) -> PublicKey:
        """Deserialize an instance from bytes."""
        if type(data) is not bytes:
            raise TypeError('cannot call from_bytes with non-bytes param')
        if len(data) < 32 or len(data) % 32 > 0:
            raise ValueError('byte length must be a multiple of 32')

        gvkey = data[:32]
        data = data[32:]
        vkeys = [gvkey]

        while len(data) > 0:
            vkeys.append(data[:32])
            data = data[32:]

        vkeys = [VerifyKey(vk) for vk in vkeys]

        return cls({'vkeys': vkeys}) if len(vkeys) > 1 else cls({'gvkey': gvkey})

    @classmethod
    def create(cls, vkeys: list[VerifyKey|bytes]) -> PublicKey:
        """Create a new PublicKey from a list or tuple of participant VerifyKeys."""
        if type(vkeys) not in (list, tuple):
            raise TypeError('vkeys must be list or tuple of VerifyKey')

        for vk in vkeys:
            if not isinstance(vk, VerifyKey):
                raise TypeError('vkeys must be list or tuple of VerifyKey')

        L = cls.encode_key_set(vkeys)
        gvkey = cls.aggregate_public_keys(vkeys, L)

        return cls({
            'vkeys': tuple(vkeys),
            'L': L,
            'gvkey': gvkey
        })

    def public(self) -> PublicKey:
        """Return a copy of the instance with only the public value (gvkey)."""
        return self.__class__({
            'gvkey': self.gvkey
        })

    def verify(self, sig: AbstractSignature) -> bool:
        """Verify a signature is valid for this PublicKey."""
        X = bytes(self.gvkey)
        s = sig.s
        R = sig.R
        M = sig.M

        c = H_sig(R, X, M)
        gs = nacl.bindings.crypto_scalarmult_ed25519_base_noclamp(s)
        RXc = nacl.bindings.crypto_scalarmult_ed25519_noclamp(c, X)
        RXc = nacl.bindings.crypto_core_ed25519_add(R, RXc)

        return bytes_are_same(gs, RXc)

    @classmethod
    def aggregate_public_keys(cls, vkeys: list[VerifyKey|bytes], key_set_L: bytes = None) -> bytes:
        """Calculate the aggregate public key from the participant keys."""
        # parse arguments
        vkeys = [vk if type(vk) is VerifyKey else VerifyKey(vk) for vk in vkeys]
        key_set_L = cls.encode_key_set(vkeys) if key_set_L is None else key_set_L

        # transform vkeys
        vkeys_transformed = [cls.pre_agg_key_transform(vk, key_set_L) for vk in vkeys]

        # sum the transformed keys
        sum = aggregate_points(vkeys_transformed)

        # return the aggregate vkey
        return sum

    @classmethod
    def pre_agg_key_transform(cls, vkey: VerifyKey, key_set_L: bytes) -> bytes:
        """Transform a participant VerifyKey prior to calculating the aggregate
            public key. This is called by aggregate_public_keys on every
            participant VerifyKey.
        """
        a_i = H_agg(key_set_L, bytes(vkey))
        return nacl.bindings.crypto_scalarmult_ed25519_noclamp(a_i, bytes(vkey))

    @classmethod
    def encode_key_set(cls, vkeys: list) -> bytes:
        """Sort the participant keys into a deterministic order, then hash the
            list to produce the keyset encoding (L).
        """
        vkeys = [bytes(vk) if type(vk) is VerifyKey else vk for vk in vkeys]
        vkeys.sort()
        return H_small(*vkeys)

    # define some properties
    @property
    def L(self) -> bytes:
        """The keyset encoding used for calculating partial signatures."""
        return self._L if hasattr(self, '_L') else None

    @L.setter
    def L(self, data: bytes):
        """The keyset encoding used for calculating partial signatures."""
        if type(data) is not bytes:
            raise TypeError('L must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('L must be bytes of len 32')

        self['L'] = data

    @property
    def gvkey(self) -> bytes:
        """The bytes of the aggregate key (denoted X in the MuSig paper)."""
        return self._gvkey if hasattr(self, '_gvkey') else None

    @gvkey.setter
    def gvkey(self, data: bytes):
        """The bytes of the aggregate key (denoted X in the MuSig paper)."""
        if type(data) is not bytes:
            raise TypeError('gvkey must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('gvkey must be bytes of len 32')

        self['gvkey'] = data

    @property
    def vkeys(self) -> tuple[VerifyKey, ...]:
        """Tuple of untransformed participant VerifyKeys."""
        return self._vkeys if hasattr(self, '_vkeys') else tuple()

    @vkeys.setter
    def vkeys(self, data: list[VerifyKey]):
        """Tuple of untransformed participant VerifyKeys."""
        if type(data) not in (list, tuple):
            raise TypeError('vkeys must be list or tuple of VerifyKeys')
        for vk in data:
            if not isinstance(vk, VerifyKey):
                raise TypeError('vkeys must be list or tuple of VerifyKeys')

        self['vkeys'] = tuple(data)



from context import musig
from json import dumps, loads
import inspect
import unittest


class TestMuSigNonceCommitment(unittest.TestCase):
    """Test suite for NonceCommitment."""
    def test_NonceCommitment_is_a_class(self):
        assert inspect.isclass(musig.NonceCommitment)

    def test_NonceCommitment_init_raises_ValueErroror_TypeError_with_empty_or_invalid_params(self):
        with self.assertRaises(TypeError):
            musig.NonceCommitment()
        with self.assertRaises(TypeError):
            musig.NonceCommitment('invalid input')
        with self.assertRaises(TypeError):
            musig.NonceCommitment(['invalid input'])

    def test_NonceCommitment_instance_has_HR_when_initialized_with_Nonce(self):
        n = musig.Nonce()
        nc = musig.NonceCommitment.create(n)
        assert hasattr(nc, 'HR')
        assert type(nc.HR) is bytes
        assert type(nc['HR']) is str

    def test_NonceCommitment_instance_is_valid_for_raises_ValueError_when_given_nonNonce(self):
        n = musig.Nonce()
        nc = musig.NonceCommitment.create(n)
        with self.assertRaises(ValueError):
            nc.is_valid_for('not a nonce')

    def test_NonceCommitment_instance_is_valid_for_method_returns_correct_values(self):
        n = musig.Nonce()
        n2 = musig.Nonce()
        nc = musig.NonceCommitment.create(n)
        assert nc.is_valid_for(n)
        assert not nc.is_valid_for(n2)

    def test_NonceCommitment_from_bytes_raises_ValueError_or_TypeError_for_invalid_serialization(self):
        with self.assertRaises(TypeError) as err:
            musig.NonceCommitment.from_bytes('not bytes')
        assert str(err.exception) == 'data must be bytes of len 32'
        with self.assertRaises(ValueError) as err:
            musig.NonceCommitment.from_bytes(b'not 32 len')
        assert str(err.exception) == 'data must be bytes of len 32'

    def test_NonceCommitment_instances_serialize_and_deserialize_properly(self):
        nc0 = musig.NonceCommitment.create(musig.Nonce())
        bts = bytes(nc0)
        str0 = str(nc0)
        js = dumps(nc0)
        nc1 = musig.NonceCommitment.from_bytes(bts)
        nc2 = musig.NonceCommitment.from_str(str0)
        nc3 = musig.NonceCommitment(loads(js))

        assert type(bts) is bytes
        assert type(str0) is str
        assert type(js) is str

        assert nc0 == nc1
        assert nc1 == nc2
        assert nc2 == nc3
        assert len(nc3.HR) == 32

    def test_NonceCommitment_instances_can_be_members_of_sets(self):
        nc0 = musig.NonceCommitment.create(musig.Nonce())
        nc1 = nc0.copy()
        nc2 = musig.NonceCommitment.create(musig.Nonce())

        ncs = list(set([nc0, nc1, nc2]))
        assert len(ncs) == 2


if __name__ == '__main__':
    unittest.main()



from musig.constants import (
    MAX_WAIT_TIME_FOR_COMMITMENTS,
    MAX_WAIT_TIME_FOR_PUBLIC_NONCES,
    MAX_WAIT_TIME_FOR_PARTIAL_SIGS,
)
from musig.helpers import (
    clamp_scalar, aggregate_points, H_big, H_small, derive_key_from_seed,
    derive_challenge, H_agg, H_sig, xor, bytes_are_same
)
from musig.nonce import Nonce
from musig.noncecommitment import NonceCommitment
from musig.partialsignature import PartialSignature
from musig.publickey import PublicKey
from musig.signature import Signature
# the below imports must be in this order due to dependency coupling
from musig.protocol import ProtocolState, ProtocolError, ProtocolMessage
from musig.singlesigkey import SingleSigKey
from musig.signingsession import SigningSession

__version__ = '0.1.0'

def version() -> str:
    """Return the version of the package."""
    return __version__



from context import musig
from json import dumps, loads
from nacl.signing import SigningKey
from time import time, sleep
from uuid import UUID, uuid4
import inspect
import musig
import unittest


class TestMuSigSigningSession(unittest.TestCase):
    """Test suite for SigningSession."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]
        cls.gvkey = 'eecb56e70d2405a849aa5e55b6e2f96aac2957dba72f8c289994c842e33ec477'
        cls.uuid = uuid4()

    def test_SigningSession_is_a_class(self):
        assert hasattr(musig, 'SigningSession')
        assert inspect.isclass(musig.SigningSession)

    def test_SigningSession__init__raises_TypeError_for_nondict_param(self):
        with self.assertRaises(TypeError) as err:
            musig.SigningSession('not a dict')
        assert str(err.exception) == 'data for initialization must be of type dict'

    def test_SigningSession__init__creates_EMPTY_instance_with_None_param(self):
        session = musig.SigningSession()
        assert isinstance(session, musig.SigningSession)
        assert hasattr(session, 'protocol_state')
        assert session.protocol_state is musig.ProtocolState.EMPTY
        assert session.id is None
        assert session.number_of_participants is None

    def test_SigningSession__init__creates_INITIALIZED_instance_with_only_skey_param(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        assert isinstance(session, musig.SigningSession)
        assert hasattr(session, 'protocol_state')
        assert session.protocol_state is musig.ProtocolState.INITIALIZED
        assert isinstance(session.id, UUID)
        assert len(session.vkeys) == 1

    def test_SigningSession__init__creates_AWAITING_PARTICIPANT_KEY_with_skey_and_number_of_participants(self):
        session = musig.SigningSession({'skey': self.signing_keys[0], 'number_of_participants': 2})
        assert isinstance(session, musig.SigningSession)
        assert hasattr(session, 'protocol_state')
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTICIPANT_KEY
        assert isinstance(session.id, UUID)
        assert len(session.vkeys) == 1

    def test_SigningSession_instances_enter_state_AWAITING_PARTICIPANT_KEY_after_setting_number_of_participants(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        assert session.protocol_state is musig.ProtocolState.INITIALIZED
        session.number_of_participants = 3
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTICIPANT_KEY

    def test_SigningSession_instance_has_add_participant_keys_method(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        assert hasattr(session, 'add_participant_keys')
        assert inspect.ismethod(session.add_participant_keys)

    def test_SigningSession_raises_ProtocolError_when_too_many_participant_keys_added(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 2
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_participant_keys(self.verify_keys)
        assert err.exception.message == 'too many participant keys added'
        assert err.exception.protocol_state == musig.ProtocolState.REJECT_PARTICIPANT_KEY
        assert session.protocol_state is musig.ProtocolState.ABORTED

    def test_SigningSession_add_participant_keys_raises_TypeError_on_invalid_param(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 3
        with self.assertRaises(TypeError) as err:
            session.add_participant_keys(b'not a VerifyKey')
        assert str(err.exception) == 'acceptable inputs are VerifyKey or list/tuple of the same'
        with self.assertRaises(TypeError) as err:
            session.add_participant_keys(['still not a VerifyKey'])

    def test_SigningSession_last_updated_does_not_change_until_protocol_changes(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 3
        assert len(session.vkeys) == 1
        assert session.public_key is None
        assert type(session.last_updated) is int
        start_time = session.last_updated
        assert session.last_updated == start_time
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTICIPANT_KEY
        session.add_participant_keys(self.verify_keys[:2])
        assert session.last_updated == start_time
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTICIPANT_KEY
        sleep(0.01)
        session.add_participant_keys(self.verify_keys[2])
        assert len(session.vkeys) == 3
        assert session.protocol_state is musig.ProtocolState.AWAITING_COMMITMENT
        assert session.last_updated != start_time

    def test_SigningSession_instances_at_state_AWAITING_COMMITMENT_have_public_key(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.signing_keys)
        assert len(session.vkeys) == 1
        assert session.public_key is None
        session.add_participant_keys(self.verify_keys)
        assert len(session.vkeys) == len(self.verify_keys)
        assert session.protocol_state is musig.ProtocolState.AWAITING_COMMITMENT
        assert hasattr(session, 'public_key')
        assert isinstance(session.public_key, musig.PublicKey)
        assert session.public_key.gvkey.hex() == self.gvkey

    def test_SigningSession_instance_has_add_nonce_commitment_method(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        assert hasattr(session, 'add_nonce_commitment')
        assert inspect.ismethod(session.add_nonce_commitment)

    def test_SigningSession_raises_PorotocolError_when_unrecognized_or_too_many_NonceCommitments_added(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 2

        with self.assertRaises(musig.ProtocolError) as err:
            session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), self.verify_keys[1])
        assert err.exception.protocol_state is musig.ProtocolState.REJECT_COMMITMENT
        assert err.exception.message == 'unrecognized vkey'
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTICIPANT_KEY

        session.add_participant_keys(self.verify_keys[:2])
        session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), self.verify_keys[0])
        session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), self.verify_keys[1])

        with self.assertRaises(musig.ProtocolError) as err:
            session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), self.verify_keys[1])
        assert err.exception.protocol_state is musig.ProtocolState.REJECT_COMMITMENT
        assert err.exception.message == 'too many nonce commitments added for this vkey'
        assert session.protocol_state is musig.ProtocolState.ABORTED

        with self.assertRaises(musig.ProtocolError) as err:
            session.number_of_participants = 1

    def test_SigningSession_raises_PorotocolError_when_max_time_elapsed_before_collecting_all_NonceCommitments(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 2
        session.add_participant_keys(self.verify_keys[:2])
        session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), self.verify_keys[0])
        # simulate having arrived at protocol_state:AWAITING_COMMITMENT in the past
        assert session.protocol_state is musig.ProtocolState.AWAITING_COMMITMENT
        session.last_updated = (time() - musig.MAX_WAIT_TIME_FOR_COMMITMENTS - 100) * 1000
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), self.verify_keys[1])
        assert err.exception.protocol_state is musig.ProtocolState.TIME_EXCEEDED_AWAITING_COMMITMENT
        assert err.exception.message == 'maximum time elapsed awaiting nonce commitments'
        assert session.protocol_state is musig.ProtocolState.ABORTED

    def test_SigningSession_enters_state_AWAITING_MESSAGE_after_receiving_all_NonceCommitments(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)
        for vk in self.verify_keys:
            session.add_nonce_commitment(musig.NonceCommitment.create(musig.Nonce()), vk)
        assert len(session.nonce_commitments) == session.number_of_participants
        assert session.protocol_state is musig.ProtocolState.AWAITING_MESSAGE

    def test_SigningSession_enters_state_AWAITING_NONCE_after_receiving_message(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)
        nonces = [musig.Nonce() for vk in self.verify_keys]
        nonces = zip(nonces, self.verify_keys)
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = zip(commitments, self.verify_keys)
        for c in commitments:
            session.add_nonce_commitment(*c)
        assert session.protocol_state is musig.ProtocolState.AWAITING_MESSAGE
        session.message = 'hello world'
        assert session.protocol_state is musig.ProtocolState.AWAITING_NONCE

    def test_SigningSession_add_nonce_raises_TypeError_or_ProtocolError_if_params_invalid(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)
        nonces = [musig.Nonce() for vk in self.verify_keys]
        nonces = zip(nonces, self.verify_keys)
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = zip(commitments, self.verify_keys)
        for c in commitments:
            session.add_nonce_commitment(c[0], c[1])
        session.message = 'hello world'

        assert session.protocol_state is musig.ProtocolState.AWAITING_NONCE
        with self.assertRaises(TypeError) as err:
            session.add_nonce('not a Nonce', self.verify_keys[0])
        with self.assertRaises(TypeError) as err:
            session.add_nonce(musig.Nonce(), 'not a VerifyKey')
        assert session.protocol_state is musig.ProtocolState.AWAITING_NONCE
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_nonce(musig.Nonce(), self.verify_keys[0])
        assert err.exception.protocol_state is musig.ProtocolState.REJECT_NONCE
        assert err.exception.message == 'Nonce invalid for NonceCommitment for this VerifyKey'
        assert session.protocol_state == musig.ProtocolState.ABORTED

    def test_SigningSession_add_nonce_raises_ProtocolError_when_max_time_elapsed_before_collecting_all_Nonces(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)
        nonces = [musig.Nonce() for vk in self.verify_keys]
        nonces = list(zip(nonces, self.verify_keys))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys))
        for c in commitments:
            session.add_nonce_commitment(*c)
        session.message = 'hello world'
        # simulate having arrived at protocol_State:AWAITING_NONCE in the past
        assert session.protocol_state is musig.ProtocolState.AWAITING_NONCE
        session.last_updated = (time() - musig.MAX_WAIT_TIME_FOR_PUBLIC_NONCES - 100) * 1000
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_nonce(*nonces[0])
        assert err.exception.protocol_state is musig.ProtocolState.TIME_EXCEEDED_AWAITING_NONCE
        assert err.exception.message == 'maximum time elapsed awaiting public nonces'
        assert session.protocol_state is musig.ProtocolState.ABORTED

    def test_SigningSession_enters_state_AWAITING_PARTIAL_SIGNATURE_after_all_Nonces_added(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)
        nonces = [musig.Nonce().public() for vk in self.verify_keys]
        nonces = list(zip(nonces, self.verify_keys))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys))
        for c in commitments:
            session.add_nonce_commitment(*c)
        session.message = 'hello world'
        assert session.protocol_state is musig.ProtocolState.AWAITING_NONCE
        for n in nonces:
            session.add_nonce(*n)
        assert len(session.nonce_points.keys()) == session.number_of_participants
        assert session.aggregate_nonce is not None
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE

    def test_SigningSession_make_partial_signature_returns_PartialSignature(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 2
        session.add_participant_keys(self.verify_keys[:2])
        nonces = [musig.Nonce() for vk in self.verify_keys[:2]]
        nonces = list(zip(nonces, self.verify_keys[:2]))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys[:2]))
        for c in commitments:
            session.add_nonce_commitment(*c)
        session.message = 'hello world'
        for n in nonces:
            session.add_nonce(*n)

        sig = session.make_partial_signature()
        assert isinstance(sig, musig.PartialSignature)

    def test_SigningSession_add_partial_signature_raises_TypeError_or_ProtocolError_if_params_invalid(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 2
        session.add_participant_keys(self.verify_keys[:2])
        nonces = [musig.Nonce() for vk in self.verify_keys[:2]]
        nonces = list(zip(nonces, self.verify_keys[:2]))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys[:2]))
        for c in commitments:
            session.add_nonce_commitment(*c)
        session.message = 'hello world'
        for n in nonces:
            session.add_nonce(*n)
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE

        sig = session.make_partial_signature()

        with self.assertRaises(TypeError):
            session.add_partial_signature('not a Partialsignature', self.verify_keys[0])
        with self.assertRaises(TypeError):
            session.add_partial_signature(sig, 'not a VerifyKey')
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_partial_signature(sig, self.verify_keys[2])
        assert err.exception.protocol_state is musig.ProtocolState.REJECT_PARTIAL_SIGNATURE
        assert err.exception.message == 'unrecognized vkey'

        # set up a copy to get a different partial signature
        another_session = musig.SigningSession({**session})
        nonce_commitments = another_session.nonce_commitments
        nonce_points = another_session.nonce_points
        n = musig.Nonce()
        nc = musig.NonceCommitment.create(n)
        nonce_commitments[another_session.skey.verify_key] = nc
        another_session.nonce_commitments = nonce_commitments
        nonce_points[another_session.skey.verify_key] = n
        another_session.nonce_points = nonce_points
        another_sig = another_session.make_partial_signature()

        # we can add the same partial signature any number of times without problem
        session.add_partial_signature(sig, self.verify_keys[0])
        session.add_partial_signature(sig, self.verify_keys[0])
        session.add_partial_signature(sig, self.verify_keys[0])

        # but an error will be raised if a conflicting signature is added
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_partial_signature(another_sig, self.verify_keys[0])
        assert err.exception.protocol_state is musig.ProtocolState.REJECT_PARTIAL_SIGNATURE
        assert err.exception.message == 'too many partial signatures added for this vkey'

    def test_SigningSession_add_partial_signatures_raises_ProtocolError_when_max_time_elapsed_before_collecting_all_PartialSignatures(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = 2
        session.add_participant_keys(self.verify_keys[:2])
        nonces = [musig.Nonce() for vk in self.verify_keys[:2]]
        nonces = list(zip(nonces, self.verify_keys[:2]))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys[:2]))
        for c in commitments:
            session.add_nonce_commitment(*c)
        session.message = 'hello world'
        for n in nonces:
            session.add_nonce(*n)
        # simulate having entered protocol_state:AWAITING_PARTIAL_SIGNATURE in the past
        session.last_updated = (time() - musig.MAX_WAIT_TIME_FOR_PARTIAL_SIGS - 100) * 1000
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE

        sig = session.make_partial_signature()
        with self.assertRaises(musig.ProtocolError) as err:
            session.add_partial_signature(sig, self.verify_keys[0])
        assert err.exception.protocol_state is musig.ProtocolState.TIME_EXCEEDED_AWAITING_PARTIAL_SIGNATURE
        assert err.exception.message == 'maximum time elapsed awaiting partial signatures'
        assert session.protocol_state is musig.ProtocolState.ABORTED

    def test_SigningSession_enters_state_COMPLETE_after_all_PartialSignatures_added_and_signature_validates(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)
        nonces = [musig.Nonce() for vk in self.verify_keys]
        nonces = list(zip(nonces, self.verify_keys))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys))
        for c in commitments:
            session.add_nonce_commitment(*c)
        session.message = 'hello world'
        for n in nonces:
            session.add_nonce(*n)
        assert session.protocol_state is musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE

        for i in range(len(self.signing_keys)):
            # NB: this works because the nonce_points are actually full nonces
            # In a realistic setting, only the nonce for the actual user will be
            # a full nonce, so hot-swapping SigningKeys won't work.
            session.skey = self.signing_keys[i]
            sig = session.make_partial_signature()
            session.add_partial_signature(sig, self.verify_keys[i])

        assert session.signature is not None
        assert session.protocol_state is musig.ProtocolState.COMPLETED

        assert session.public_key.verify(session.signature)

    def test_SigningSession_instances_serialize_and_deserialize_from_json(self):
        session = musig.SigningSession({'skey': self.signing_keys[0]})
        session.number_of_participants = len(self.verify_keys)
        session.add_participant_keys(self.verify_keys)

        serialized = dumps(session)
        deserialized = musig.SigningSession(loads(serialized))
        assert deserialized.id == session.id
        assert deserialized.protocol_state == session.protocol_state
        assert deserialized.number_of_participants == session.number_of_participants
        assert deserialized.last_updated == session.last_updated
        assert deserialized.skey == session.skey
        assert deserialized.vkeys == session.vkeys
        assert deserialized.nonce_commitments == session.nonce_commitments
        assert deserialized.nonce_points == session.nonce_points
        assert deserialized.aggregate_nonce == session.aggregate_nonce
        assert deserialized.message == session.message
        assert deserialized.partial_signatures == session.partial_signatures
        assert deserialized.public_key == session.public_key
        assert deserialized.signature == session.signature

        nonces = [musig.Nonce() for vk in self.verify_keys]
        nonces = list(zip(nonces, self.verify_keys))
        commitments = [musig.NonceCommitment.create(n[0]) for n in nonces]
        commitments = list(zip(commitments, self.verify_keys))
        for c in commitments:
            session.add_nonce_commitment(*c)

        serialized = dumps(session)
        deserialized = musig.SigningSession(loads(serialized))
        assert deserialized.id == session.id
        assert deserialized.protocol_state == session.protocol_state
        assert deserialized.number_of_participants == session.number_of_participants
        assert deserialized.last_updated == session.last_updated
        assert deserialized.skey == session.skey
        assert deserialized.vkeys == session.vkeys
        assert deserialized.nonce_commitments == session.nonce_commitments
        assert deserialized.nonce_points == session.nonce_points
        assert deserialized.aggregate_nonce == session.aggregate_nonce
        assert deserialized.message == session.message
        assert deserialized.partial_signatures == session.partial_signatures
        assert deserialized.public_key == session.public_key
        assert deserialized.signature == session.signature

        session.message = 'hello world'
        for n in nonces:
            session.add_nonce(*n)

        serialized = dumps(session)
        deserialized = musig.SigningSession(loads(serialized))
        assert deserialized.id == session.id
        assert deserialized.protocol_state == session.protocol_state
        assert deserialized.number_of_participants == session.number_of_participants
        assert deserialized.last_updated == session.last_updated
        assert deserialized.skey == session.skey
        assert deserialized.vkeys == session.vkeys
        assert deserialized.nonce_commitments == session.nonce_commitments
        assert deserialized.nonce_points == session.nonce_points
        assert deserialized.aggregate_nonce == session.aggregate_nonce
        assert deserialized.message == session.message
        assert deserialized.partial_signatures == session.partial_signatures
        assert deserialized.public_key == session.public_key
        assert deserialized.signature == session.signature

        for i in range(len(self.signing_keys)):
            session.skey = self.signing_keys[i]
            sig = session.make_partial_signature()
            session.add_partial_signature(sig, self.verify_keys[i])

        serialized = dumps(session)
        deserialized = musig.SigningSession(loads(serialized))
        assert deserialized.id == session.id
        assert deserialized.protocol_state == session.protocol_state
        assert deserialized.number_of_participants == session.number_of_participants
        assert deserialized.last_updated == session.last_updated
        assert deserialized.skey == session.skey
        assert deserialized.vkeys == session.vkeys
        assert deserialized.nonce_commitments == session.nonce_commitments
        assert deserialized.nonce_points == session.nonce_points
        assert deserialized.aggregate_nonce == session.aggregate_nonce
        assert deserialized.message == session.message
        assert deserialized.partial_signatures == session.partial_signatures
        assert deserialized.public_key == session.public_key
        assert deserialized.signature == session.signature


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from base64 import b64decode
from json import dumps, loads
from musig.abstractclasses import AbstractSigningSession
from musig.constants import (
    MAX_WAIT_TIME_FOR_COMMITMENTS,
    MAX_WAIT_TIME_FOR_PARTIAL_SIGS,
    MAX_WAIT_TIME_FOR_PUBLIC_NONCES,
)
from musig.helpers import aggregate_points
from musig.nonce import Nonce
from musig.noncecommitment import NonceCommitment
from musig.partialsignature import PartialSignature
from musig.protocol import ProtocolState, ProtocolError
from musig.publickey import PublicKey
from musig.signature import Signature
from nacl.signing import SigningKey, VerifyKey
from time import time
from uuid import UUID, uuid4


class SigningSession(AbstractSigningSession):
    """A class that handles multi-party signing sessions.
        This is designed to maintain security with a 3-round protocol.
        Though the keys are reusable, each Session can be used for only a single
        signature to avoid certain cryptographic attacks. Nonce commitments
        (hash(R_i)) are pre-shared instead of public nonces (R_i) to avoid
        vulnerability to the Wagner attack. Detailed documentation can be found
        in the docs/musig.md file (eventually).
    """

    def __init__(self, data: dict = None) -> None:
        """Init method. Initialize with None to create an EMPTY SigningSession.
            Initialize with {'skey': SigningKey} to create an INITIALIZED SigningSession.
            Initialize with a dict (output from json serialization/deserialization)
            to restore a previously used or customly configured SigningSession.
        """
        self.last_updated = int(time() * 1000)

        if data is None:
            self.protocol_state = ProtocolState.EMPTY
            return

        if data is not None and type(data) is not dict:
            raise TypeError('data for initialization must be of type dict')

        if 'skey' in data:
            if type(data['skey']) is SigningKey:
                self.skey = data['skey']
            else:
                seed = data['skey'] if type(data['skey']) is bytes else b64decode(data['skey'])
                self.skey = SigningKey(seed)

            if 'id' not in data:
                # create an INITIALIZED session
                self.id = uuid4()
                self.vkeys = (self.skey.verify_key,)
                self.protocol_state = ProtocolState.INITIALIZED

        if 'id' in data:
            sid = data['id'] if type(data['id']) is bytes else b64decode(data['id'])
            self.id = UUID(bytes=sid)

        if 'protocol_state' in data:
            if type(data['protocol_state']) is ProtocolState:
                self.protocol_state = data['protocol_state']
            else:
                self.protocol_state = ProtocolState[data['protocol_state']]

        if 'number_of_participants' in data:
            self.number_of_participants = data['number_of_participants']

        if 'last_updated' in data:
            self.last_updated = data['last_updated']

        if 'vkeys' in data:
            vkeys = [vk if type(vk) in (bytes, VerifyKey) else b64decode(vk) for vk in data['vkeys']]
            self.vkeys = tuple([vk if type(vk) is VerifyKey else VerifyKey(vk) for vk in vkeys])

        if 'nonce_commitments' in data:
            commitments = {}
            for name in data['nonce_commitments']:
                n = VerifyKey(name if type(name) is bytes else b64decode(name))
                c = data['nonce_commitments'][name]
                c = c if type(c) is bytes else b64decode(c)
                commitments[n] = NonceCommitment.from_bytes(c)
            self.nonce_commitments = commitments

        if 'nonce_points' in data:
            nonce_points = {}
            for name in data['nonce_points']:
                n = VerifyKey(name if type(name) is bytes else b64decode(name))
                c = data['nonce_points'][name]
                c = c if type(c) is bytes else b64decode(c)
                nonce_points[n] = Nonce.from_bytes(c)
            self.nonce_points = nonce_points

        if 'aggregate_nonce' in data:
            n = data['aggregate_nonce']
            self.aggregate_nonce = Nonce.from_bytes(n if type(n) is bytes else b64decode(n))

        if 'message' in data:
            m = data['message']
            self.message = m if type(m) is bytes else b64decode(m)

        if 'partial_signatures' in data:
            parts = {}
            for name in data['partial_signatures']:
                n = VerifyKey(name if type(name) is bytes else b64decode(name))
                ps = data['partial_signatures'][name]
                parts[n] = PartialSignature.from_bytes(ps if type(ps) is bytes else b64decode(ps))
            self.partial_signatures = parts

        if 'public_key' in data:
            pk = data['public_key']
            self.public_key = PublicKey.from_bytes(pk if type(pk) is bytes else b64decode(pk))

        if 'signature' in data:
            s = data['signature']
            self.signature = Signature.from_bytes(s if type(s) is bytes else b64decode(s))

    def __setitem__(self, key, value) -> None:
        """Slightly modified __setitem__ for updating timestamps appropriately."""
        super().__setitem__(key, value)

        if key == 'protocol_state':
            self._protocol_state = value if type(value) is ProtocolState else ProtocolState[value]
            self.last_updated = int(time() * 1000)
        elif key == 'last_updated':
            ...
        else:
            self.update_protocol_state()

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on instance; i.e. serialize to bytes."""
        return bytes(dumps(self), 'utf-8')

    @classmethod
    def from_bytes(cls, data: bytes) -> SigningSession:
        """Deserializes output from __bytes__."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes')
        return SigningSession(loads(str(data, 'utf-8')))

    def add_participant_keys(self, keys: VerifyKey|list[VerifyKey]) -> None:
        """Add participant VerifyKey(s)."""
        if isinstance(keys, VerifyKey):
            keys = [keys]

        if type(keys) not in (list, tuple):
            raise TypeError('acceptable inputs are VerifyKey or list/tuple of the same')

        vkeys = list(self.vkeys)
        for k in keys:
            if not isinstance(k, VerifyKey):
                raise TypeError('acceptable inputs are VerifyKey or list/tuple of the same')
            if k not in self.vkeys:
                vkeys.append(k)

        self.vkeys = tuple(vkeys)

        if len(self.vkeys) == self.number_of_participants:
            # if all participant keys have been gathered, derive the aggregate public key
            self.public_key = PublicKey.create(list(self.vkeys))

    def add_nonce_commitment(self, commitment: NonceCommitment, vkey: VerifyKey) -> None:
        """Add a NonceCommitment from a participant identified by the VerifyKey."""
        if not isinstance(commitment, NonceCommitment):
            raise TypeError('commitment must be NonceCommitment')
        if not isinstance(vkey, VerifyKey):
            raise TypeError('vkey must be a VerifyKey')
        if vkey not in self.vkeys:
            raise ProtocolError('unrecognized vkey', ProtocolState.REJECT_COMMITMENT)
        if vkey in self.nonce_commitments.keys():
            self.protocol_state = ProtocolState.ABORTED
            raise ProtocolError('too many nonce commitments added for this vkey', ProtocolState.REJECT_COMMITMENT)

        # update in this round-about way to make use of setter and __setitem__ logic
        nonce_commitments = self.nonce_commitments
        nonce_commitments[vkey] = commitment

        self.nonce_commitments = nonce_commitments

    def add_nonce(self, nonce: Nonce, vkey: VerifyKey) -> None:
        """Add a Nonce from a participant identified by the VerifyKey."""
        if not isinstance(nonce, Nonce):
            raise TypeError('nonce must be Nonce')
        if not isinstance(vkey, VerifyKey):
            raise TypeError('vkey must be a VerifyKey')
        if vkey not in self.vkeys:
            raise ProtocolError('unrecognized vkey', ProtocolState.REJECT_NONCE)
        if vkey in self.nonce_points.keys():
            self.protocol_state = ProtocolState.ABORTED
            raise ProtocolError('too many nonces added for this vkey', ProtocolState.REJECT_NONCE)
        if vkey not in self.nonce_commitments.keys():
            raise ProtocolError('no NonceCommitment for this vkey', ProtocolState.REJECT_NONCE)

        nonce_points = self.nonce_points
        if not self.nonce_commitments[vkey].is_valid_for(nonce):
            # abort if the nonce does not validate for its commitment
            self.protocol_state = ProtocolState.ABORTED
            raise ProtocolError('Nonce invalid for NonceCommitment for this VerifyKey', ProtocolState.REJECT_NONCE)

        # update in this round-about way to make use of setter and __setitem__ logic
        nonce_points[vkey] = nonce
        self.nonce_points = nonce_points

    def make_partial_signature(self) -> PartialSignature:
        """Create a partial signature to be broadcast to other participants."""
        return PartialSignature.create(self.skey, self.nonce_points[self.skey.verify_key].r,
            self.public_key.L, self.public_key, self.aggregate_nonce.R, self.message)

    def add_partial_signature(self, sig: PartialSignature, vkey: VerifyKey) -> None:
        """Add a PartialSignature from a participant identified by the VerifyKey."""
        if not isinstance(sig, PartialSignature):
            raise TypeError('sig must be a PartialSignature')
        if not isinstance(vkey, VerifyKey):
            raise TypeError('vkey must be a VerifyKey')
        if vkey not in self.vkeys:
            raise ProtocolError('unrecognized vkey', ProtocolState.REJECT_PARTIAL_SIGNATURE)
        if vkey in self.partial_signatures.keys() and sig != self.partial_signatures[vkey]:
            self.protocol_state = ProtocolState.ABORTED
            raise ProtocolError('too many partial signatures added for this vkey', ProtocolState.REJECT_PARTIAL_SIGNATURE)

        # update in this round-about way to make use of setter and __setitem__ logic
        partial_signatures = self.partial_signatures
        partial_signatures[vkey] = sig
        self.partial_signatures = partial_signatures

    def update_protocol_state(self) -> None:
        """Handle transitions between ProtocolStates as the SigningSession values
            are updated. This is called automatically after any value is updated.
            The protocol state will only update when the necessary conditions
            have been met.
        """
        elapsed_time = int((time() * 1000 - self.last_updated)/1000)

        # abort conditions from misuse
        if self.number_of_participants is not None:
            if len(self.vkeys) > self.number_of_participants:
                self.protocol_state = ProtocolState.ABORTED
                raise ProtocolError('too many participant keys added', ProtocolState.REJECT_PARTICIPANT_KEY)
            if len(self.nonce_commitments.keys()) > self.number_of_participants:
                self.protocol_state = ProtocolState.ABORTED
                raise ProtocolError('too many nonce commitments added', ProtocolState.REJECT_COMMITMENT)

        if self.protocol_state is ProtocolState.EMPTY:
            if self.id is not None and self.skey is not None:
                self.protocol_state = ProtocolState.INITIALIZED

        if self.protocol_state is ProtocolState.INITIALIZED:
            if self.number_of_participants is not None and len(self.vkeys) < self.number_of_participants:
                self.protocol_state = ProtocolState.AWAITING_PARTICIPANT_KEY
            elif self.vkeys is not None and len(self.vkeys) == self.number_of_participants:
                self.protocol_state = ProtocolState.AWAITING_COMMITMENT

        if self.protocol_state is ProtocolState.AWAITING_PARTICIPANT_KEY:
            if self.vkeys is not None and len(self.vkeys) == self.number_of_participants:
                self.protocol_state = ProtocolState.AWAITING_COMMITMENT

        if self.protocol_state is ProtocolState.AWAITING_COMMITMENT:
            if elapsed_time > MAX_WAIT_TIME_FOR_COMMITMENTS:
                self.protocol_state = ProtocolState.ABORTED
                raise ProtocolError('maximum time elapsed awaiting nonce commitments', ProtocolState.TIME_EXCEEDED_AWAITING_COMMITMENT)
            if self.number_of_participants is not None:
                if len(self.nonce_commitments.keys()) == self.number_of_participants:
                    self.protocol_state = ProtocolState.AWAITING_MESSAGE

        if self.protocol_state is ProtocolState.AWAITING_MESSAGE:
            if self.message is not None:
                self.protocol_state = ProtocolState.AWAITING_NONCE

        if self.protocol_state is ProtocolState.AWAITING_NONCE:
            if elapsed_time > MAX_WAIT_TIME_FOR_PUBLIC_NONCES:
                self.protocol_state = ProtocolState.ABORTED
                raise ProtocolError('maximum time elapsed awaiting public nonces', ProtocolState.TIME_EXCEEDED_AWAITING_NONCE)
            if self.aggregate_nonce is not None:
                self.protocol_state = ProtocolState.AWAITING_PARTIAL_SIGNATURE
            elif self.number_of_participants is not None:
                if len(self.nonce_points.keys()) == self.number_of_participants:
                    points = [self.nonce_points[vk].R for vk in self.nonce_points]
                    self.aggregate_nonce = Nonce.from_bytes(aggregate_points(points))

        if self.protocol_state is ProtocolState.AWAITING_PARTIAL_SIGNATURE:
            if elapsed_time > MAX_WAIT_TIME_FOR_PARTIAL_SIGS:
                self.protocol_state = ProtocolState.ABORTED
                raise ProtocolError('maximum time elapsed awaiting partial signatures', ProtocolState.TIME_EXCEEDED_AWAITING_PARTIAL_SIGNATURE)
            if self.signature is not None:
                self.protocol_state = ProtocolState.COMPLETED
            elif self.number_of_participants is not None:
                if len(self.partial_signatures.keys()) == self.number_of_participants:
                    parts = [self.partial_signatures[vk] for vk in self.partial_signatures]
                    self.signature = Signature.create(self.aggregate_nonce.R, self.message, parts)

    @property
    def id(self) -> UUID|None:
        """The UUID of the session."""
        return self._id if hasattr(self, '_id') else None

    @id.setter
    def id(self, value: UUID):
        """The UUID of the session."""
        if not isinstance(value, UUID):
            raise TypeError('id must be a UUID')

        self['id'] = value.bytes
        self._id = value

    @property
    def number_of_participants(self) -> int|None:
        """The number of participants expected to participate in the protocol."""
        return self._number_of_participants if hasattr(self, '_number_of_participants') else None

    @number_of_participants.setter
    def number_of_participants(self, value: int):
        """The number of participants expected to participate in the protocol."""
        if not isinstance(value, int):
            raise TypeError('number_of_participants must be an int')

        self['number_of_participants'] = value

    @property
    def protocol_state(self) -> ProtocolState|None:
        """The current state of the session."""
        return self._protocol_state if hasattr(self, '_protocol_state') else None

    @protocol_state.setter
    def protocol_state(self, value: ProtocolState):
        """The current state of the session."""
        if not isinstance(value, ProtocolState):
            raise TypeError('protocol_state must be a ProtocolState')

        self['protocol_state'] = value.name

    @property
    def last_updated(self) -> int|None:
        """A timestamp recording the last time the protocol state was updated."""
        return self._last_updated if hasattr(self, '_last_updated') else None

    @last_updated.setter
    def last_updated(self, value: int):
        """A timestamp recording the last time the protocol state was updated."""
        if type(value) not in (float, int):
            raise TypeError('last_updated must be a timestamp')

        self['last_updated'] = int(value)

    @property
    def skey(self) -> SigningKey|None:
        """The SigningKey of the participant using this instance."""
        return self._skey if hasattr(self, '_skey') else None

    @skey.setter
    def skey(self, value: SigningKey):
        """The SigningKey of the participant using this instance."""
        if not isinstance(value, SigningKey):
            raise TypeError('skey must be a nacl.signing.SigningKey')

        self['skey'] = value

    @property
    def vkeys(self) -> tuple[VerifyKey, ...]:
        """A tuple of participant VerifyKeys."""
        return self._vkeys if hasattr(self, '_vkeys') else tuple()

    @vkeys.setter
    def vkeys(self, value: tuple[VerifyKey, ...]):
        """A tuple of participant VerifyKeys."""
        if type(value) not in (tuple, list):
            raise TypeError('vkeys must be a tuple or list of nacl.signing.VerifyKeys')

        for vk in value:
            if not isinstance(vk, VerifyKey):
                raise TypeError('vkeys must be a tuple or list of nacl.signing.VerifyKeys')

        self['vkeys'] = tuple(value)

    @property
    def nonce_commitments(self) -> dict[VerifyKey, NonceCommitment]:
        """A dict mapping participant VerifyKey to NonceCommitment."""
        return self._nonce_commitments if hasattr(self, '_nonce_commitments') else dict()

    @nonce_commitments.setter
    def nonce_commitments(self, value: dict[VerifyKey, NonceCommitment]):
        """A dict mapping participant VerifyKey to NonceCommitment."""
        if not isinstance(value, dict):
            raise TypeError('nonce_commitments must be a dict of form {VerifyKey:NonceCommitment}')

        for vk in value:
            nc = value[vk]
            if not isinstance(vk, VerifyKey):
                raise TypeError('nonce_commitments must be a dict of form {VerifyKey:NonceCommitment}')
            if not isinstance(nc, NonceCommitment):
                raise TypeError('nonce_commitments must be a dict of form {VerifyKey:NonceCommitment}')

        self['nonce_commitments'] = value

    @property
    def nonce_points(self) -> dict[VerifyKey, Nonce]:
        """A dict mapping participant VerifyKey to Nonce. Note that the Nonce
            for the participant using this instance will include the private
            scalar value, but the Nonces of other participants will include only
            the public point values.
        """
        return self._nonce_points if hasattr(self, '_nonce_points') else dict()

    @nonce_points.setter
    def nonce_points(self, value: dict[VerifyKey, Nonce]):
        """A dict mapping participant VerifyKey to Nonce. Note that the Nonce
            for the participant using this instance will include the private
            scalar value, but the Nonces of other participants will include only
            the public point values.
        """
        if not isinstance(value, dict):
            raise TypeError('nonce_points must be a dict of form {VerifyKey:Nonce}')

        for vk in value:
            n = value[vk]
            if not isinstance(vk, VerifyKey):
                raise TypeError('nonce_points must be a dict of form {VerifyKey:Nonce}')
            if not isinstance(n, Nonce):
                raise TypeError('nonce_points must be a dict of form {VerifyKey:Nonce}')

        self['nonce_points'] = value

    @property
    def aggregate_nonce(self) -> Nonce|None:
        """The aggregate nonce point for the session."""
        return self._aggregate_nonce if hasattr(self, '_aggregate_nonce') else None

    @aggregate_nonce.setter
    def aggregate_nonce(self, value: Nonce):
        """The aggregate nonce point for the session."""
        if not isinstance(value, Nonce):
            raise TypeError('aggregate_nonce must be a Nonce')

        self['aggregate_nonce'] = value

    @property
    def message(self) -> bytes|None:
        """The message to be n-of-n signed."""
        return self._message if hasattr(self, '_message') else None

    @message.setter
    def message(self, value: bytes|str):
        """The message to be n-of-n signed."""
        value = bytes(value, 'utf-8') if type(value) is str else value

        if type(value) is not bytes:
            raise TypeError('message must be bytes or str')

        self['message'] = value

    @property
    def partial_signatures(self) -> dict[VerifyKey, PartialSignature]:
        """A dict mapping participant VerifyKey to PartialSignature (public values s_i only)."""
        return self._partial_signatures if hasattr(self, '_partial_signatures') else dict()

    @partial_signatures.setter
    def partial_signatures(self, value: dict[VerifyKey, PartialSignature]):
        """A dict mapping participant VerifyKey to PartialSignature (public values s_i only)."""
        if not isinstance(value, dict):
            raise TypeError('partial_signatures must be a dict of form {VerifyKey:PartialSignature}')

        for vk in value:
            ps = value[vk]
            if not isinstance(vk, VerifyKey):
                raise TypeError('partial_signatures must be a dict of form {VerifyKey:PartialSignature}')
            if not isinstance(ps, PartialSignature):
                raise TypeError('partial_signatures must be a dict of form {VerifyKey:PartialSignature}')

        self['partial_signatures'] = value

    @property
    def public_key(self) -> PublicKey|None:
        """The aggregate public key for the session."""
        return self._public_key if hasattr(self, '_public_key') else None

    @public_key.setter
    def public_key(self, value: PublicKey):
        """The aggregate public key for the session."""
        if not isinstance(value, PublicKey):
            raise TypeError('public_key must be a PublicKey')

        self['public_key'] = value

    @property
    def signature(self) -> Signature|None:
        """The final n-of-n signature."""
        return self._signature if hasattr(self, '_signature') else None

    @signature.setter
    def signature(self, value: Signature):
        """The final n-of-n signature."""
        if not isinstance(value, Signature):
            raise TypeError('signature must be a Signature')

        self['signature'] = value



from __future__ import annotations
from base64 import b64decode
from musig.abstractclasses import AbstractNonce, AbstractNonceCommitment
from musig.helpers import bytes_are_same, H_small


class NonceCommitment(AbstractNonceCommitment):
    """A class that handles generating, serializing, and deserializing nonce
        commitments.
    """

    def __init__(self, data: dict) -> None:
        """Initialize with a dict."""
        if type(data) is not dict:
            raise TypeError('data must be type dict')

        if 'HR' in data:
            HR = data['HR']
            self.HR = HR if type(HR) is bytes else b64decode(HR)

    def __bytes__(self) -> bytes:
        """Result of calling bytes() on an instance."""
        if self.HR is not None:
            return self.HR
        return b''

    @classmethod
    def create(cls, nonce: AbstractNonce) -> NonceCommitment:
        """Create a new instance by hashing the given nonce."""
        if not isinstance(nonce, AbstractNonce):
            raise TypeError('nonce must be instance of AbstractNonce')

        HR = H_small(nonce.R)
        return cls({'HR': HR})

    def copy(self) -> NonceCommitment:
        """Make a copy without serializing and deserializing."""
        return NonceCommitment({**self})

    def is_valid_for(self, nonce: AbstractNonce) -> bool:
        """Checks if the NonceCommitment is valid for a specific Nonce.
            Comparison is done via xor'ing bytes to avoid timing attacks.
        """
        if not isinstance(nonce, AbstractNonce):
            raise ValueError('supplied nonce must be an instance of Nonce')

        return bytes_are_same(H_small(nonce.R), self.HR)

    @classmethod
    def from_bytes(cls, data: bytes) -> NonceCommitment:
        """Deserializes output from __bytes__."""
        if type(data) is not bytes:
            raise TypeError('data must be bytes of len 32')
        if len(data) != 32:
            raise ValueError('data must be bytes of len 32')

        return cls({'HR': data})

    @property
    def HR(self) -> bytes|None:
        """The nonce commitment bytes (hash of a nonce point)."""
        return self._HR if hasattr(self, '_HR') else None

    @HR.setter
    def HR(self, value: bytes):
        """The nonce commitment bytes (hash of a nonce point)."""
        if type(value) is not bytes:
            raise TypeError('HR must be bytes')
        if len(value) != 32:
            raise ValueError('HR value must have length 32')

        self['HR'] = value



from context import musig
import unittest


class TestMuSigConstants(unittest.TestCase):
    """Test suite for the constants used by the various classes."""

    def test_musig_has_MAX_WAIT_TIME_FOR_COMMITMENTS(self):
        assert hasattr(musig, 'MAX_WAIT_TIME_FOR_COMMITMENTS')
        assert type(musig.MAX_WAIT_TIME_FOR_COMMITMENTS) is int

    def test_musig_has_MAX_WAIT_TIME_FOR_PUBLIC_NONCES(self):
        assert hasattr(musig, 'MAX_WAIT_TIME_FOR_PUBLIC_NONCES')
        assert type(musig.MAX_WAIT_TIME_FOR_PUBLIC_NONCES) is int

    def test_musig_has_MAX_WAIT_TIME_FOR_PARTIAL_SIGS(self):
        assert hasattr(musig, 'MAX_WAIT_TIME_FOR_PARTIAL_SIGS')
        assert type(musig.MAX_WAIT_TIME_FOR_PARTIAL_SIGS) is int


if __name__ == '__main__':
    unittest.main()



"""1-of-1 example"""


from musig import SingleSigKey
from nacl.signing import SigningKey, VerifyKey, SignedMessage
from secrets import token_bytes


# do something to load a seed for private key creation, e.g.
seed = token_bytes()
skey = SigningKey(seed)
ssk = SingleSigKey({'skey': skey})

# sign a message and distribute along with public key for verification
message = b'hello world'
sig = ssk.sign_message(message)
pubkey = ssk.vkey

# to verify the signature, use the pubkey
assert pubkey.verify(sig)

# alternatively use a VerifyKey
vkey = VerifyKey(bytes(pubkey.public()))
assert vkey.verify(SignedMessage(bytes(sig))) == message

# call str on any object to get the hex representation
# call repr on any object to get the jsonable dict representation
print(f'{str(sig)=}')
print(f'{repr(sig)=}\n')
print(f'{str(pubkey)=}')
print(f'{repr(pubkey)=}')



from setuptools import setup, find_packages

with open('readme.md', 'r') as f:
    readme = f.read()

with open('license.txt', 'r') as f:
    license = f.read()

with open('requirements.txt', 'r') as f:
    requirements = f.read()

setup(
    name='MuSig',
    version='0.0.1',
    author='k98kurz@github',
    url='https://github.com/k98kurz/musig',
    description='Simple-to-use package implementing a MuSig multi-sig protocol.',
    long_description=readme,
    license=license,
    packages=find_packages(exclude=('tests', 'docs', 'examples')),
    install_requires=requirements,
    python_requires='>=3.6.0',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Security :: Cryptography",
        "License :: OSI Approved :: ISC License (ISCL)",
    ],
)